#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import enum


# a.k.a. SHT (section header type)
@enum.unique
class Elf32SectionHeaderType(enum.Enum):
    NULL          =  0
    PROGBITS      =  1
    SYMTAB        =  2
    STRTAB        =  3
    RELA          =  4
    HASH          =  5
    DYNAMIC       =  6
    NOTE          =  7
    NOBITS        =  8
    REL           =  9

    MIPS_GPTAB    = 0x70000003
    MIPS_DEBUG    = 0x70000005
    MIPS_REGINFO  = 0x70000006
    MIPS_OPTIONS  = 0x7000000D
    MIPS_ABIFLAGS = 0x7000002A


# a.k.a. STT (symbol table type)
@enum.unique
class Elf32SymbolTableType(enum.Enum):
    NOTYPE       =  0
    OBJECT       =  1
    FUNC         =  2
    SECTION      =  3
    FILE         =  4
    COMMON       =  5
    TLS          =  6
    NUM          =  7



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from .. import symbols

from . import SectionBase


class SectionData(SectionBase):
    def __init__(self, context: common.Context, vram: int|None, filename: str, array_of_bytes: bytearray):
        super().__init__(context, vram, filename, array_of_bytes, common.FileSectionType.Data)


    def analyze(self):
        self.checkAndCreateFirstSymbol()

        symbolList: list[tuple[int, int, str]] = []
        localOffset = 0

        for w in self.words:
            currentVram = self.getVramOffset(localOffset)

            contextSym = self.context.getSymbol(currentVram, tryPlusOffset=False)
            if contextSym is not None:
                contextSym.isDefined = True
                symbolList.append((localOffset, currentVram, contextSym.name))

            if self.vram is not None:
                if w >= self.vram and w < 0x84000000:
                    if self.context.getAnySymbol(w) is None:
                        self.context.newPointersInData.add(w)

            localOffset += 4

        for i, (offset, vram, symName) in enumerate(symbolList):
            if i + 1 == len(symbolList):
                words = self.words[offset//4:]
            else:
                nextOffset = symbolList[i+1][0]
                words = self.words[offset//4:nextOffset//4]

            sym = symbols.SymbolData(self.context, offset + self.inFileOffset, vram, symName, words)
            sym.setCommentOffset(self.commentOffset)
            sym.analyze()
            self.symbolList.append(sym)


    def removePointers(self) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = False
        for i in range(self.sizew):
            top_byte = (self.words[i] >> 24) & 0xFF
            if top_byte == 0x80:
                self.words[i] = top_byte << 24
                was_updated = True
            if (top_byte & 0xF0) == 0x00 and (top_byte & 0x0F) != 0x00:
                self.words[i] = top_byte << 24
                was_updated = True

        return was_updated



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionVectorId, InstructionNormal, instructionDescriptorDict
from .MipsInstructionConfig import InstructionConfig


class InstructionNormalRsp(InstructionNormal):
    RemovedOpcodes: dict[int, InstructionId] = {
        0b010_100: InstructionId.BEQL,
        0b010_101: InstructionId.BNEL,
        0b010_110: InstructionId.BLEZL,
        0b010_111: InstructionId.BGTZL,

        0b011_000: InstructionId.DADDI,
        0b011_001: InstructionId.DADDIU,
        0b011_010: InstructionId.LDL,
        0b011_011: InstructionId.LDR,

        0b100_010: InstructionId.LWL,
        0b100_110: InstructionId.LWR,
        0b100_111: InstructionId.LWU,

        0b101_010: InstructionId.SWL,
        0b101_100: InstructionId.SDL,
        0b101_101: InstructionId.SDR,
        0b101_110: InstructionId.SWR,

        0b110_000: InstructionId.LL,
        0b110_010: InstructionId.LWC2,
        0b110_100: InstructionId.LLD,
        0b110_101: InstructionId.LDC1,
        0b110_110: InstructionId.LDC2,
        0b110_111: InstructionId.LD,

        0b111_000: InstructionId.SC,
        0b111_010: InstructionId.SWC2,
        0b111_100: InstructionId.SCD,
        0b111_101: InstructionId.SDC1,
        0b111_110: InstructionId.SDC2,
        0b111_111: InstructionId.SD,
    }
    Opcodes_BySWC2: dict[int, InstructionVectorId] = {
        0b00_000: InstructionVectorId.SBV,
        0b00_001: InstructionVectorId.SSV,
        0b00_010: InstructionVectorId.SLV,
        0b00_011: InstructionVectorId.SDV,

        0b00_100: InstructionVectorId.SQV,
        0b00_101: InstructionVectorId.SRV,

        0b00_110: InstructionVectorId.SPV,

        # 0b00_111: InstructionVectorId.SUV,
        0b00_111: InstructionVectorId.SWV,

        0b01_000: InstructionVectorId.SHV,
        0b01_001: InstructionVectorId.SFV,

        0b01_011: InstructionVectorId.STV,
    }
    Opcodes_ByLWC2: dict[int, InstructionVectorId] = {
        0b00_000: InstructionVectorId.LBV,
        0b00_001: InstructionVectorId.LSV,
        0b00_010: InstructionVectorId.LLV,
        0b00_011: InstructionVectorId.LDV,

        0b00_100: InstructionVectorId.LQV,
        0b00_101: InstructionVectorId.LRV,

        0b00_110: InstructionVectorId.LPV,

        0b00_111: InstructionVectorId.LUV,

        0b01_000: InstructionVectorId.LHV,
        0b01_001: InstructionVectorId.LFV,

        0b01_011: InstructionVectorId.LTV,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        if self.opcode not in self.RemovedOpcodes:
            self.uniqueId = self.NormalOpcodes.get(self.opcode, InstructionId.INVALID)

        # SWC2
        if self.opcode == 0b111_010:
            if self.rd in self.Opcodes_BySWC2:
                self.uniqueId = self.Opcodes_BySWC2[self.rd]
                if self.elementLow == 0:
                    if self.uniqueId == InstructionVectorId.SWV:
                        self.uniqueId = InstructionVectorId.SUV
        # LWC2
        elif self.opcode == 0b110_010:
            if self.rd in self.Opcodes_ByLWC2:
                self.uniqueId = self.Opcodes_ByLWC2[self.rd]

        if self.uniqueId in instructionDescriptorDict:
            self.descriptor = instructionDescriptorDict[self.uniqueId]


    def getRegisterName(self, register: int) -> str:
        return self.getGprRspRegisterName(register)


    def disassembleInstruction(self, immOverride: str|None=None) -> str:
        opcode = self.getOpcodeName()
        formated_opcode = opcode.lower().ljust(InstructionConfig.OPCODE_LJUST + self.extraLjustWidthOpcode, ' ')
        vt = self.getVectorRspRegisterName(self.vt)
        base = self.getGprRspRegisterName(self.baseRegister)
        offset = hex(self.offsetVector)
        element = self.processVectorElement(self.elementLow)

        if self.uniqueId in (InstructionVectorId.LSV, InstructionVectorId.SSV, ):
            offset = hex(self.offsetVector << 1)
        elif self.uniqueId in (InstructionVectorId.LLV, InstructionVectorId.SLV, ):
            offset = hex(self.offsetVector << 2)
        elif self.uniqueId in (InstructionVectorId.LDV, InstructionVectorId.SDV,
                               InstructionVectorId.LPV, InstructionVectorId.SPV,
                               InstructionVectorId.LUV, InstructionVectorId.SUV, ):
            offset = hex(self.offsetVector << 3)
        elif self.uniqueId in (InstructionVectorId.LQV, InstructionVectorId.SQV,
                               InstructionVectorId.LRV, InstructionVectorId.SRV,
                               InstructionVectorId.LHV, InstructionVectorId.SHV,
                               InstructionVectorId.LFV, InstructionVectorId.SFV,
                               InstructionVectorId.LTV, InstructionVectorId.STV, InstructionVectorId.SWV, ):
            offset = hex(self.offsetVector << 4)

        result = f"{formated_opcode} "
        # SWC2, LWC2
        if self.opcode in (0b111_010, 0b110_010, ):
            result += f"{vt}[{element}],"
            result = result.ljust(14, ' ')
            result += f" {offset}({base})"

            return result

        return super().disassembleInstruction(immOverride)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .Elf32Constants import Elf32SectionHeaderType, Elf32SymbolTableType
from .Elf32Header import Elf32Header
from .Elf32SectionHeaders import Elf32SectionHeaders, Elf32SectionHeaderEntry
from .Elf32StringTable import Elf32StringTable
from .Elf32Syms import Elf32Syms, Elf32SymEntry
from .Elf32Rels import Elf32Rels, Elf32RelEntry

from .Elf32File import Elf32File



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionBase, instructionDescriptorDict
from .MipsInstructionConfig import InstructionConfig


class InstructionCoprocessor1(InstructionBase):
    Cop1Opcodes_ByFormat = {
        0b00_000: InstructionId.MFC1,
        0b00_001: InstructionId.DMFC1,
        0b00_010: InstructionId.CFC1,

        0b00_100: InstructionId.MTC1,
        0b00_101: InstructionId.DMTC1,
        0b00_110: InstructionId.CTC1,
    }
    Cop1Opcodes_ByFunction = {
        0b000_000: { 0: InstructionId.ADD_S, 1: InstructionId.ADD_D },
        0b000_001: { 0: InstructionId.SUB_S, 1: InstructionId.SUB_D },
        0b000_010: { 0: InstructionId.MUL_S, 1: InstructionId.MUL_D },
        0b000_011: { 0: InstructionId.DIV_S, 1: InstructionId.DIV_D },

        0b000_100: { 0: InstructionId.SQRT_S, 1: InstructionId.SQRT_D },
        0b000_101: { 0: InstructionId.ABS_S,  1: InstructionId.ABS_D },
        0b000_110: { 0: InstructionId.MOV_S,  1: InstructionId.MOV_D },
        0b000_111: { 0: InstructionId.NEG_S,  1: InstructionId.NEG_D },

        0b001_000: { 0: InstructionId.ROUND_L_S, 1: InstructionId.ROUND_L_D },
        0b001_001: { 0: InstructionId.TRUNC_L_S, 1: InstructionId.TRUNC_L_D },
        0b001_010: { 0: InstructionId.CEIL_L_S, 1: InstructionId.CEIL_L_D },
        0b001_011: { 0: InstructionId.FLOOR_L_S, 1: InstructionId.FLOOR_L_D },

        0b001_100: { 0: InstructionId.ROUND_W_S, 1: InstructionId.ROUND_W_D },
        0b001_101: { 0: InstructionId.TRUNC_W_S, 1: InstructionId.TRUNC_W_D },
        0b001_110: { 0: InstructionId.CEIL_W_S, 1: InstructionId.CEIL_W_D },
        0b001_111: { 0: InstructionId.FLOOR_W_S, 1: InstructionId.FLOOR_W_D },
    }
    CompareConditionsCodes = {
        0b0_000: { 0: InstructionId.C_F_S,    1: InstructionId.C_F_D }, # False
        0b0_001: { 0: InstructionId.C_UN_S,   1: InstructionId.C_UN_D }, # UNordered
        0b0_010: { 0: InstructionId.C_EQ_S,   1: InstructionId.C_EQ_D }, # EQual
        0b0_011: { 0: InstructionId.C_UEQ_S,  1: InstructionId.C_UEQ_D }, # Unordered or EQual
        0b0_100: { 0: InstructionId.C_OLT_S,  1: InstructionId.C_OLT_D }, # Ordered or Less Than
        0b0_101: { 0: InstructionId.C_ULT_S,  1: InstructionId.C_ULT_D }, # Unordered or Less Than
        0b0_110: { 0: InstructionId.C_OLE_S,  1: InstructionId.C_OLE_D }, # Ordered or Less than or Equal
        0b0_111: { 0: InstructionId.C_ULE_S,  1: InstructionId.C_ULE_D }, # Unordered or Less than or Equal

        0b1_000: { 0: InstructionId.C_SF_S,   1: InstructionId.C_SF_D }, # Signaling False
        0b1_001: { 0: InstructionId.C_NGLE_S, 1: InstructionId.C_NGLE_D }, # Not Greater than or Less than or Equal
        0b1_010: { 0: InstructionId.C_SEQ_S,  1: InstructionId.C_SEQ_D }, # Signaling Equal
        0b1_011: { 0: InstructionId.C_NGL_S,  1: InstructionId.C_NGL_D }, # Not Greater than or Less than
        0b1_100: { 0: InstructionId.C_LT_S,   1: InstructionId.C_LT_D }, # Less than
        0b1_101: { 0: InstructionId.C_NGE_S,  1: InstructionId.C_NGE_D }, # Not Greater than or Equal
        0b1_110: { 0: InstructionId.C_LE_S,   1: InstructionId.C_LE_D }, # Less than or Equal
        0b1_111: { 0: InstructionId.C_NGT_S,  1: InstructionId.C_NGT_D }, # Not Greater than
    }
    ConvertCodes = {
        0b000: { 0b001: InstructionId.CVT_S_D, 0b100: InstructionId.CVT_S_W, 0b101: InstructionId.CVT_S_L },
        0b001: { 0b000: InstructionId.CVT_D_S, 0b100: InstructionId.CVT_D_W, 0b101: InstructionId.CVT_D_L },
        0b100: { 0b000: InstructionId.CVT_W_S, 0b001: InstructionId.CVT_W_D, },
        0b101: { 0b000: InstructionId.CVT_L_S, 0b001: InstructionId.CVT_L_D, },
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()


    def processUniqueId(self):
        if self.fmt in InstructionCoprocessor1.Cop1Opcodes_ByFormat:
            self.uniqueId = InstructionCoprocessor1.Cop1Opcodes_ByFormat[self.fmt]

        elif self.fmt == 0b01_000: # fmt = BC
            tf = (self.instr >> 16) & 0x01
            nd = (self.instr >> 17) & 0x01
            if tf:
                if nd:
                    self.uniqueId = InstructionId.BC1TL
                else:
                    self.uniqueId = InstructionId.BC1T
            else:
                if nd:
                    self.uniqueId = InstructionId.BC1FL
                else:
                    self.uniqueId = InstructionId.BC1F

        elif self.function in InstructionCoprocessor1.Cop1Opcodes_ByFunction:
            perFmt = InstructionCoprocessor1.Cop1Opcodes_ByFunction[self.function]
            fmt = self.fmt & 0x07
            if fmt in perFmt:
                self.uniqueId = perFmt[fmt]

        elif self.fc == 0b11:
            if self.cond in InstructionCoprocessor1.CompareConditionsCodes:
                perFmt = InstructionCoprocessor1.CompareConditionsCodes[self.cond]
                fmt = self.fmt & 0x07
                if fmt in perFmt:
                    self.uniqueId = perFmt[fmt]

        elif self.fc == 0b10:
            fun = self.function & 0x07
            if fun in InstructionCoprocessor1.ConvertCodes:
                perFmt = InstructionCoprocessor1.ConvertCodes[fun]
                fmt = self.fmt & 0x07
                if fmt in perFmt:
                    self.uniqueId = perFmt[fmt]

        self.descriptor = instructionDescriptorDict[self.uniqueId]

    def blankOut(self):
        if self.fmt in InstructionCoprocessor1.Cop1Opcodes_ByFormat:
            self.rt = 0
            self.rd = 0
            self.sa = 0
            self.function = 0
        elif self.fmt == 0b01_000: # fmt = BC
            self.rd = 0
            self.sa = 0
            self.function = 0
        elif self.function in InstructionCoprocessor1.Cop1Opcodes_ByFunction:
            self.rt = 0
            self.rd = 0
            self.sa = 0
        elif self.fc == 0b11 or self.fc == 0b10:
            self.rt = 0
            self.rd = 0
            self.sa = 0

    def getOpcodeName(self) -> str:
        return super().getOpcodeName().replace("_", ".")



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from . import InstructionId, InstructionBase, instructionDescriptorDict
from .MipsInstructionConfig import InstructionConfig


class InstructionNormal(InstructionBase):
    NormalOpcodes: dict[int, InstructionId] = {
        # 0b000_000: "SPECIAL",
        # 0b000_001: "REGIMM",
        0b000_010: InstructionId.J,
        0b000_011: InstructionId.JAL,
        0b000_100: InstructionId.BEQ,
        0b000_101: InstructionId.BNE,
        0b000_110: InstructionId.BLEZ,
        0b000_111: InstructionId.BGTZ,

        0b001_000: InstructionId.ADDI,
        0b001_001: InstructionId.ADDIU,
        0b001_010: InstructionId.SLTI,
        0b001_011: InstructionId.SLTIU,
        0b001_100: InstructionId.ANDI,
        0b001_101: InstructionId.ORI,
        0b001_110: InstructionId.XORI,
        0b001_111: InstructionId.LUI,

        # 0b010_000: "COP0", # Coprocessor OPeration z
        # 0b010_001: "COP1", # Coprocessor OPeration z
        # 0b010_010: "COP2", # Coprocessor OPeration z
        # 0b010_011: "COP3", # Coprocessor OPeration z
        0b010_100: InstructionId.BEQL,
        0b010_101: InstructionId.BNEL,
        0b010_110: InstructionId.BLEZL,
        0b010_111: InstructionId.BGTZL,

        0b011_000: InstructionId.DADDI,
        0b011_001: InstructionId.DADDIU,
        0b011_010: InstructionId.LDL,
        0b011_011: InstructionId.LDR,
        # 0b011_100: "",
        # 0b011_101: "",
        # 0b011_110: "",
        # 0b011_111: "",

        0b100_000: InstructionId.LB,
        0b100_001: InstructionId.LH,
        0b100_010: InstructionId.LWL,
        0b100_011: InstructionId.LW,
        0b100_100: InstructionId.LBU,
        0b100_101: InstructionId.LHU,
        0b100_110: InstructionId.LWR,
        0b100_111: InstructionId.LWU,

        0b101_000: InstructionId.SB,
        0b101_001: InstructionId.SH,
        0b101_010: InstructionId.SWL,
        0b101_011: InstructionId.SW,
        0b101_100: InstructionId.SDL,
        0b101_101: InstructionId.SDR,
        0b101_110: InstructionId.SWR,
        0b101_111: InstructionId.CACHE,

        0b110_000: InstructionId.LL,
        0b110_001: InstructionId.LWC1,
        0b110_010: InstructionId.LWC2,
        0b110_011: InstructionId.PREF,
        0b110_100: InstructionId.LLD,
        0b110_101: InstructionId.LDC1,
        0b110_110: InstructionId.LDC2,
        0b110_111: InstructionId.LD,

        0b111_000: InstructionId.SC,
        0b111_001: InstructionId.SWC1,
        0b111_010: InstructionId.SWC2,
        # 0b111_011: "",
        0b111_100: InstructionId.SCD,
        0b111_101: InstructionId.SDC1,
        0b111_110: InstructionId.SDC2,
        0b111_111: InstructionId.SD,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()


    def processUniqueId(self):
        self.uniqueId = self.NormalOpcodes.get(self.opcode, InstructionId.INVALID)

        if InstructionConfig.PSEUDO_INSTRUCTIONS:
            if self.rt == 0:
                if self.uniqueId == InstructionId.BEQ:
                    if self.rs == 0:
                        if InstructionConfig.PSEUDO_B:
                            self.uniqueId = InstructionId.B
                    else:
                        if InstructionConfig.PSEUDO_BEQZ:
                            self.uniqueId = InstructionId.BEQZ
                elif self.uniqueId == InstructionId.BNE:
                    if InstructionConfig.PSEUDO_BNEZ:
                        self.uniqueId = InstructionId.BNEZ

        self.descriptor = instructionDescriptorDict[self.uniqueId]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from . import SymbolBase


class SymbolText(SymbolBase):
    def __init__(self, context: common.Context, inFileOffset: int, vram: int|None, name: str, words: list[int]):
        super().__init__(context, inFileOffset, vram, name, words, common.FileSectionType.Text)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from ..MipsFileBase import FileBase

class SectionBase(FileBase):
    def blankOutDifferences(self, other: FileBase) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = False
        if len(common.GlobalConfig.IGNORE_WORD_LIST) > 0:
            min_len = min(self.sizew, other.sizew)
            for i in range(min_len):
                for upperByte in common.GlobalConfig.IGNORE_WORD_LIST:
                    word = upperByte << 24
                    if ((self.words[i] >> 24) & 0xFF) == upperByte and ((other.words[i] >> 24) & 0xFF) == upperByte:
                        self.words[i] = word
                        other.words[i] = word
                        was_updated = True

        return was_updated



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .MipsInstructionConfig import InstructionConfig
from .MipsConstants import InstructionId, instructionDescriptorDict, InstrType
from .MipsInstructionBase import InstructionBase


class InstructionSpecial(InstructionBase):
    SpecialOpcodes: dict[int, InstructionId] = {
        0b000_000: InstructionId.SLL,
        # 0b000_001: "MOVCI", # TODO
        0b000_010: InstructionId.SRL,
        0b000_011: InstructionId.SRA,
        0b000_100: InstructionId.SLLV,
        # 0b000_101: "",
        0b000_110: InstructionId.SRLV,
        0b000_111: InstructionId.SRAV,

        0b001_000: InstructionId.JR,
        0b001_001: InstructionId.JALR,
        0b001_010: InstructionId.MOVZ,
        0b001_011: InstructionId.MOVN,
        0b001_100: InstructionId.SYSCALL,
        0b001_101: InstructionId.BREAK,
        # 0b001_110: "",
        0b001_111: InstructionId.SYNC,

        0b010_000: InstructionId.MFHI,
        0b010_001: InstructionId.MTHI,
        0b010_010: InstructionId.MFLO,
        0b010_011: InstructionId.MTLO,
        0b010_100: InstructionId.DSLLV,
        # 0b010_101: "",
        0b010_110: InstructionId.DSRLV,
        0b010_111: InstructionId.DSRAV,

        0b011_000: InstructionId.MULT,
        0b011_001: InstructionId.MULTU,
        0b011_010: InstructionId.DIV,
        0b011_011: InstructionId.DIVU,
        0b011_100: InstructionId.DMULT,
        0b011_101: InstructionId.DMULTU,
        0b011_110: InstructionId.DDIV,
        0b011_111: InstructionId.DDIVU,

        0b100_000: InstructionId.ADD,
        0b100_001: InstructionId.ADDU,
        0b100_010: InstructionId.SUB,
        0b100_011: InstructionId.SUBU,
        0b100_100: InstructionId.AND,
        0b100_101: InstructionId.OR,
        0b100_110: InstructionId.XOR,
        0b100_111: InstructionId.NOR,

        # 0b101_000: "",
        # 0b101_001: "",
        0b101_010: InstructionId.SLT,
        0b101_011: InstructionId.SLTU,
        0b101_100: InstructionId.DADD,
        0b101_101: InstructionId.DADDU,
        0b101_110: InstructionId.DSUB,
        0b101_111: InstructionId.DSUBU,

        0b110_000: InstructionId.TGE,
        0b110_001: InstructionId.TGEU,
        0b110_010: InstructionId.TLT,
        0b110_011: InstructionId.TLTU,
        0b110_100: InstructionId.TEQ,
        # 0b110_101: "",
        0b110_110: InstructionId.TNE,
        # 0b110_111: "",

        0b111_000: InstructionId.DSLL,
        # 0b111_001: "",
        0b111_010: InstructionId.DSRL,
        0b111_011: InstructionId.DSRA,
        0b111_100: InstructionId.DSLL32,
        # 0b111_101: "",
        0b111_110: InstructionId.DSRL32,
        0b111_111: InstructionId.DSRA32,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()


    def processUniqueId(self):
        self.uniqueId = self.SpecialOpcodes.get(self.function, InstructionId.INVALID)

        if self.instr == 0:
            # NOP is special enough
            self.uniqueId = InstructionId.NOP
        elif InstructionConfig.PSEUDO_INSTRUCTIONS:
            if self.rt == 0:
                if self.uniqueId == InstructionId.OR:
                    if InstructionConfig.PSEUDO_MOVE:
                        self.uniqueId = InstructionId.MOVE
                elif self.uniqueId == InstructionId.NOR:
                    if InstructionConfig.PSEUDO_NOT:
                        self.uniqueId = InstructionId.NOT
            elif self.uniqueId == InstructionId.SUBU:
                if self.rs == 0:
                    if InstructionConfig.PSEUDO_NEGU:
                        self.uniqueId = InstructionId.NEGU

        self.descriptor = instructionDescriptorDict[self.uniqueId]

        if self.uniqueId == InstructionId.JALR:
            # $ra
            if self.rd != 31:
                self.descriptor = instructionDescriptorDict[InstructionId.JALR_RD]


    def blankOut(self):
        self.rs = 0
        self.rt = 0
        self.rd = 0
        self.sa = 0


    def disassembleInstruction(self, immOverride: str|None=None) -> str:
        patch = False

        if InstructionConfig.SN64_DIV_FIX:
            if self.uniqueId == InstructionId.BREAK:
                patch = True
            elif self.uniqueId == InstructionId.DIV:
                if not self.inHandwrittenFunction:
                    self.descriptor = instructionDescriptorDict[InstructionId.SN64_DIV]
            elif self.uniqueId == InstructionId.DIVU:
                if not self.inHandwrittenFunction:
                    self.descriptor = instructionDescriptorDict[InstructionId.SN64_DIVU]

        if self.descriptor.instrType == InstrType.typeR and "code" not in self.descriptor.operands:
            if "rs" not in self.descriptor.operands and self.rs != 0:
                patch = True
            if "rt" not in self.descriptor.operands and self.rt != 0:
                patch = True
            if "rd" not in self.descriptor.operands and self.rd != 0 and self.uniqueId != InstructionId.JALR:
                patch = True
            if "sa" not in self.descriptor.operands and self.sa != 0:
                patch = True

        if patch:
            patchedResult = self.disassembleAsData()
            patchedResult += " # "
            patchedResult += super().disassembleInstruction(immOverride)
            return patchedResult

        return super().disassembleInstruction(immOverride)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionId, InstructionSpecial

from .MipsInstructionConfig import InstructionConfig
from .MipsConstants import InstructionId, instructionDescriptorDict


class InstructionSpecialRsp(InstructionSpecial):
    RemovedOpcodes: dict[int, InstructionId] = {
        0b001_100: InstructionId.SYSCALL,
        0b001_111: InstructionId.SYNC,

        0b010_000: InstructionId.MFHI,
        0b010_001: InstructionId.MTHI,
        0b010_010: InstructionId.MFLO,
        0b010_011: InstructionId.MTLO,
        0b010_100: InstructionId.DSLLV,
        0b010_110: InstructionId.DSRLV,
        0b010_111: InstructionId.DSRAV,

        0b011_000: InstructionId.MULT,
        0b011_001: InstructionId.MULTU,
        0b011_010: InstructionId.DIV,
        0b011_011: InstructionId.DIVU,
        0b011_100: InstructionId.DMULT,
        0b011_101: InstructionId.DMULTU,
        0b011_110: InstructionId.DDIV,
        0b011_111: InstructionId.DDIVU,

        0b101_100: InstructionId.DADD,
        0b101_101: InstructionId.DADDU,
        0b101_110: InstructionId.DSUB,
        0b101_111: InstructionId.DSUBU,

        0b110_000: InstructionId.TGE,
        0b110_001: InstructionId.TGEU,
        0b110_010: InstructionId.TLT,
        0b110_011: InstructionId.TLTU,
        0b110_100: InstructionId.TEQ,
        0b110_110: InstructionId.TNE,

        0b111_000: InstructionId.DSLL,
        0b111_010: InstructionId.DSRL,
        0b111_011: InstructionId.DSRA,
        0b111_100: InstructionId.DSLL32,
        0b111_110: InstructionId.DSRL32,
        0b111_111: InstructionId.DSRA32,
    }

    def __init__(self, instr: int):
        super().__init__(instr)

        self.processUniqueId()
        self._handwrittenCategory = True


    def processUniqueId(self):
        if self.function not in self.RemovedOpcodes:
            self.uniqueId = self.SpecialOpcodes.get(self.function, InstructionId.INVALID)

        if InstructionConfig.PSEUDO_INSTRUCTIONS:
            if self.instr == 0:
                self.uniqueId = InstructionId.NOP
            elif self.rt == 0:
                if self.uniqueId == InstructionId.OR:
                    self.uniqueId = InstructionId.MOVE
                elif self.uniqueId == InstructionId.NOR:
                    self.uniqueId = InstructionId.NOT
            elif self.uniqueId == InstructionId.SUBU:
                if self.rs == 0:
                    self.uniqueId = InstructionId.NEGU

        self.descriptor = instructionDescriptorDict[self.uniqueId]

        if self.uniqueId == InstructionId.JALR:
            # $ra
            if self.rd != 31:
                self.descriptor = instructionDescriptorDict[InstructionId.JALR_RD]

        if InstructionConfig.SN64_DIV_FIX:
            if self.uniqueId in (InstructionId.DIV, InstructionId.DIVU):
                self.descriptor.operands = ["rs", "rt"]


    def getRegisterName(self, register: int) -> str:
        return self.getGprRspRegisterName(register)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from . import SymbolBase


class SymbolRodata(SymbolBase):
    def __init__(self, context: common.Context, inFileOffset: int, vram: int|None, name: str, words: list[int]):
        super().__init__(context, inFileOffset, vram, name, words, common.FileSectionType.Rodata)

        assert self.contextSym is not None
        self.contextSym: common.ContextSymbol


    def isString(self) -> bool:
        return self.contextSym.isString()

    def isFloat(self, index: int) -> bool:
        if self.contextSym.isFloat():
            word = self.words[index]
            # Filter out NaN and infinity
            if (word & 0x7F800000) != 0x7F800000:
                return True
        return False

    def isDouble(self, index: int) -> bool:
        if self.contextSym.isDouble():
            if index + 1 < self.sizew:
                word0 = self.words[index]
                word1 = self.words[index+1]
                # Filter out NaN and infinity
                if (((word0 << 32) | word1) & 0x7FF0000000000000) != 0x7FF0000000000000:
                    # Prevent accidentally losing symbols
                    currentVram = self.getVramOffset(index*4)
                    if self.context.getGenericSymbol(currentVram+4, False) is None:
                        return True
        return False

    def isJumpTable(self) -> bool:
        return self.contextSym.isJumpTable()


    def isRdata(self) -> bool:
        "Checks if the current symbol is .rdata"
        if self.contextSym is None:
            return False

        if self.contextSym.isMaybeConstVariable():
            return True

        # This symbol could be an unreferenced non-const variable
        if self.contextSym.referenceCounter == 1:
            # This const variable was already used in a function
            return False

        return True


    def renameBasedOnType(self):
        if not common.GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_DATA_TYPE:
            return

        if self.vram is None:
            return

        if not self.contextSym.isAutogenerated:
            return

        if not self.isJumpTable():
            if self.isFloat(0):
                self.contextSym.name = f"FLT_{self.vram:08X}"
            elif self.isDouble(0):
                self.contextSym.name = f"DBL_{self.vram:08X}"
            elif self.isString():
                self.contextSym.name = f"STR_{self.vram:08X}"
        self.name = self.contextSym.name

    def analyze(self):
        if self.contextSym.isDouble():
            if self.sizew % 2 != 0:
                # doubles require an even amount of words
                self.contextSym.type = None
            else:
                for i in range(self.sizew // 2):
                    if not self.isDouble(i*2):
                        # checks there's no other overlaping symbols
                        self.contextSym.type = None
                        break

        super().analyze()

    def getNthWord(self, i: int, canReferenceSymbolsWithAddends: bool=False, canReferenceConstants: bool=False) -> tuple[str, int]:
        if self.contextSym is not None:
            if self.contextSym.isByte() or self.contextSym.isShort():
                return super().getNthWord(i, canReferenceSymbolsWithAddends, canReferenceConstants)

        localOffset = 4*i
        w = self.words[i]

        label = ""
        rodataWord: int|None = w
        value: str = f"0x{w:08X}"

        # try to get the symbol name from the offset of the file (possibly from a .o elf file)
        possibleSymbolName = self.context.getOffsetGenericSymbol(self.inFileOffset + localOffset, self.sectionType)
        if possibleSymbolName is not None:
            label = possibleSymbolName.getSymbolLabel() + common.GlobalConfig.LINE_ENDS

        if len(self.context.relocSymbols[self.sectionType]) > 0:
            possibleReference = self.context.getRelocSymbol(self.inFileOffset + localOffset, self.sectionType)
            if possibleReference is not None:
                value = possibleReference.getNamePlusOffset(w)
                if possibleReference.type == common.SymbolSpecialType.jumptablelabel:
                    if w in self.context.offsetJumpTablesLabels:
                        value = self.context.offsetJumpTablesLabels[w].getName()

        dotType = ".word"
        skip = 0

        if self.isFloat(i):
            dotType = ".float"
            value = str(common.Utils.wordToFloat(w))
        elif self.isDouble(i):
            dotType = ".double"
            otherHalf = self.words[i+1]
            doubleWord = (w << 32) | otherHalf
            value = str(common.Utils.qwordToDouble(doubleWord))
            rodataWord = doubleWord
            skip = 1
        else:
            labelSym = self.context.getGenericLabel(w)
            if labelSym is not None:
                value = labelSym.getName()
            elif self.isString():
                try:
                    buffer = bytearray(4*len(self.words))
                    common.Utils.beWordsToBytes(self.words, buffer)
                    decodedValue, rawStringSize = common.Utils.decodeString(buffer, 4*i)
                    dotType = ".asciz"
                    value = f'"{decodedValue}"'
                    value += common.GlobalConfig.LINE_ENDS + (22 * " ") + ".balign 4"
                    rodataWord = None
                    skip = rawStringSize // 4
                except (UnicodeDecodeError, RuntimeError):
                    # Not a string
                    pass

        comment = self.generateAsmLineComment(localOffset, rodataWord)
        return f"{label}{comment} {dotType} {value}" + common.GlobalConfig.LINE_ENDS, skip



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import enum


@enum.unique
class InputEndian(enum.Enum):
    BIG = enum.auto()
    LITTLE = enum.auto()
    MIDDLE = enum.auto()


class GlobalConfig:
    DISASSEMBLE_UNKNOWN_INSTRUCTIONS: bool = False
    """Try to disassemble non implemented instructions and functions"""

    ADD_NEW_SYMBOLS: bool = True
    PRODUCE_SYMBOLS_PLUS_OFFSET: bool = True
    TRUST_USER_FUNCTIONS: bool = True
    TRUST_JAL_FUNCTIONS: bool = True
    TREAT_J_AS_UNCONDITIONAL_BRANCH: bool = False

    STRING_GUESSER: bool = True
    """Rodata string guesser"""

    AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE: bool = True
    """Name autogenerated symbols after the section those are come from

    Use R_ for symbols in .rodata and B_ for symbols in .bss"""

    AUTOGENERATED_NAMES_BASED_ON_DATA_TYPE: bool = True
    """Name autogenerated symbols after their type

    Use STR_ for strings, FLT_ for floats and DBL_ for doubles"""

    ENDIAN: InputEndian = InputEndian.BIG
    """Endian for input binary files"""

    GP_VALUE: int|None = None
    """Value used for $gp relocation loads and stores"""

    SYMBOL_FINDER_FILTER_LOW_ADDRESSES: bool = True
    """Toggle pointer detection for lower addresses (lower than 0x40000000)"""
    SYMBOL_FINDER_FILTER_HIGH_ADDRESSES: bool = True
    """Toggle pointer detection for higher addresses (higher than 0xC0000000)"""
    SYMBOL_FINDER_FILTERED_ADDRESSES_AS_CONSTANTS: bool = True
    """Treat filtered out addresses as constants pairs"""


    ASM_COMMENT: bool = True
    """Toggle the comments in generated assembly code"""
    GLABEL_ASM_COUNT: bool = True
    """Toggle the glabel count comment on functions"""

    ASM_TEXT_LABEL: str = "glabel"
    ASM_DATA_LABEL: str = "glabel"
    ASM_TEXT_ENT_LABEL: str = ""
    ASM_TEXT_END_LABEL: str = ""
    ASM_TEXT_FUNC_AS_LABEL: bool = False

    PRINT_NEW_FILE_BOUNDARIES: bool = False
    """Print to stdout every file boundary found in .text and .rodata"""

    USE_DOT_BYTE: bool = True
    """Disassemble symbols marked as bytes with .byte instead of .word"""
    USE_DOT_SHORT: bool = True
    """Disassemble symbols marked as shorts with .short instead of .word"""

    LINE_ENDS: str = "\n"


    QUIET: bool = False
    VERBOSE: bool = False


    PRINT_FUNCTION_ANALYSIS_DEBUG_INFO: bool = False
    PRINT_SYMBOL_FINDER_DEBUG_INFO: bool = False
    PRINT_UNPAIRED_LUIS_DEBUG_INFO: bool = False


    REMOVE_POINTERS: bool = False
    IGNORE_BRANCHES: bool = False
    """Ignores the address of every branch, jump and jal"""
    IGNORE_WORD_LIST: set[int] = set()
    """Ignores words that starts in 0xXX"""
    WRITE_BINARY: bool = False
    """write to files splitted binaries"""


    @staticmethod
    def addParametersToArgParse(parser: argparse.ArgumentParser):
        backendConfig = parser.add_argument_group("Disassembler backend configuration")

        backendConfig.add_argument("--disasm-unknown", help=f"Force disassembling functions with unknown instructions. Defaults to {GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS}", action=argparse.BooleanOptionalAction)

        backendConfig.add_argument("--j-branch", help=f"Treat J instructions as unconditional branches. {GlobalConfig.TREAT_J_AS_UNCONDITIONAL_BRANCH}", action=argparse.BooleanOptionalAction)

        backendConfig.add_argument("--string-guesser", help=f"Toggles the string guesser feature. Defaults to {GlobalConfig.STRING_GUESSER}", action=argparse.BooleanOptionalAction)

        backendConfig.add_argument("--name-vars-by-section", help=f"Toggles the naming-after-section feature for autogenerated names. This means autogenerated symbols get a R_ or B_ prefix if the symbol is from a rodata or bss section. Defaults to {GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE}", action=argparse.BooleanOptionalAction)
        backendConfig.add_argument("--name-vars-by-type", help=f"Toggles the naming-after-type feature for autogenerated names. This means autogenerated symbols can get a STR_, FLT_ or DBL_ prefix if the symbol is a string, float or double. Defaults to {GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_DATA_TYPE}", action=argparse.BooleanOptionalAction)

        backendConfig.add_argument("--endian", help=f"Set the endianness of input files. Defaults to {GlobalConfig.ENDIAN.name.lower()}", choices=["big", "little", "middle"], default=GlobalConfig.ENDIAN.name.lower())

        backendConfig.add_argument("--gp", help="Set the value used for loads and stores related to the $gp register. A hex value is expected")

        backendConfig.add_argument("--filter-low-addresses", help=f"Filter out low addresses (lower than 0x40000000) when searching for pointers. Defaults to {GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES}", action=argparse.BooleanOptionalAction)
        backendConfig.add_argument("--filter-high-addresses", help=f"Filter out high addresses (higher than 0xC0000000) when searching for pointers. Defaults to {GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES}", action=argparse.BooleanOptionalAction)
        backendConfig.add_argument("--filtered-addresses-as-constants", help=f"Treat filtered out addressed as constants. Defaults to {GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_CONSTANTS}", action=argparse.BooleanOptionalAction)


        miscConfig = parser.add_argument_group("Disassembler misc options")

        miscConfig.add_argument("--asm-comments", help=f"Toggle the comments in generated assembly code. Defaults to {GlobalConfig.ASM_COMMENT}", action=argparse.BooleanOptionalAction)
        miscConfig.add_argument("--glabel-count", help=f"Toggle glabel count comment. Defaults to {GlobalConfig.GLABEL_ASM_COUNT}", action=argparse.BooleanOptionalAction)

        miscConfig.add_argument("--asm-text-label", help=f"Changes the label used to declare functions. Defaults to {GlobalConfig.ASM_TEXT_LABEL}")
        miscConfig.add_argument("--asm-data-label", help=f"Changes the label used to declare data symbols. Defaults to {GlobalConfig.ASM_DATA_LABEL}")
        miscConfig.add_argument("--asm-ent-label", help=f"Tells the disassembler to start using an ent label for functions")
        miscConfig.add_argument("--asm-end-label", help=f"Tells the disassembler to start using an end label for functions")
        miscConfig.add_argument("--asm-func-as-label", help=f"Toggle adding the function name as an additional label. Defaults to {GlobalConfig.ASM_TEXT_FUNC_AS_LABEL}", action=argparse.BooleanOptionalAction)

        miscConfig.add_argument("--print-new-file-boundaries", help=f"Print to stdout any new file boundary found. Defaults to {GlobalConfig.PRINT_NEW_FILE_BOUNDARIES}", action=argparse.BooleanOptionalAction)

        miscConfig.add_argument("--use-dot-byte", help=f"Disassemble symbols marked as bytes with .byte instead of .word. Defaults to {GlobalConfig.USE_DOT_BYTE}", action=argparse.BooleanOptionalAction)
        miscConfig.add_argument("--use-dot-short", help=f"Disassemble symbols marked as shorts with .short instead of .word. Defaults to {GlobalConfig.USE_DOT_SHORT}", action=argparse.BooleanOptionalAction)


        verbosityConfig = parser.add_argument_group("Verbosity options")

        verbosityConfig.add_argument("-V", "--verbose", help="Enable verbose mode", action=argparse.BooleanOptionalAction)
        verbosityConfig.add_argument("-q", "--quiet", help="Silence most of the output", action=argparse.BooleanOptionalAction)


        debugging = parser.add_argument_group("Disassembler debugging options")

        debugging.add_argument("--debug-func-analysis", help="Enables some debug info printing related to the function analysis)", action=argparse.BooleanOptionalAction)
        debugging.add_argument("--debug-symbol-finder", help="Enables some debug info printing related to the symbol finder system)", action=argparse.BooleanOptionalAction)
        debugging.add_argument("--debug-unpaired-luis", help="Enables some debug info printing related to the unpaired LUI instructions)", action=argparse.BooleanOptionalAction)


    @classmethod
    def parseArgs(cls, args: argparse.Namespace):
        if args.disasm_unknown is not None:
            GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS = args.disasm_unknown

        if args.j_branch is not None:
            GlobalConfig.TREAT_J_AS_UNCONDITIONAL_BRANCH = args.j_branch

        if args.string_guesser is not None:
            GlobalConfig.STRING_GUESSER = args.string_guesser

        if args.name_vars_by_section is not None:
            GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE = args.name_vars_by_section
        if args.name_vars_by_type is not None:
            GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_DATA_TYPE = args.name_vars_by_type

        if args.endian == "little":
            GlobalConfig.ENDIAN = InputEndian.LITTLE
        elif args.endian == "middle":
            GlobalConfig.ENDIAN = InputEndian.MIDDLE
        else:
            GlobalConfig.ENDIAN = InputEndian.BIG

        if args.gp is not None:
            GlobalConfig.GP_VALUE = int(args.gp, 16)

        if args.filter_low_addresses is not None:
            GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES = args.filter_low_addresses
        if args.filter_high_addresses is not None:
            GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES = args.filter_high_addresses
        if args.filtered_addresses_as_constants is not None:
            GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_CONSTANTS = args.filtered_addresses_as_constants


        if args.asm_comments is not None:
            GlobalConfig.ASM_COMMENT = args.asm_comments
        if args.glabel_count is not None:
            GlobalConfig.GLABEL_ASM_COUNT = args.glabel_count

        if args.asm_text_label:
            GlobalConfig.ASM_TEXT_LABEL = args.asm_text_label
        if args.asm_data_label:
            GlobalConfig.ASM_DATA_LABEL = args.asm_data_label
        if args.asm_ent_label:
            GlobalConfig.ASM_TEXT_ENT_LABEL = args.asm_ent_label
        if args.asm_end_label:
            GlobalConfig.ASM_TEXT_END_LABEL = args.asm_end_label
        if args.asm_func_as_label is not None:
            GlobalConfig.ASM_TEXT_FUNC_AS_LABEL = args.asm_func_as_label

        if args.print_new_file_boundaries is not None:
            GlobalConfig.PRINT_NEW_FILE_BOUNDARIES = args.print_new_file_boundaries

        if args.use_dot_byte is not None:
            GlobalConfig.USE_DOT_BYTE = args.use_dot_byte
        if args.use_dot_short is not None:
            GlobalConfig.USE_DOT_SHORT = args.use_dot_short


        if args.verbose is not None:
            GlobalConfig.VERBOSE = args.verbose
        if args.quiet is not None:
            GlobalConfig.QUIET = args.quiet


        if args.debug_func_analysis is not None:
            GlobalConfig.PRINT_FUNCTION_ANALYSIS_DEBUG_INFO = args.debug_func_analysis
        if args.debug_symbol_finder is not None:
            GlobalConfig.PRINT_SYMBOL_FINDER_DEBUG_INFO = args.debug_symbol_finder
        if args.debug_unpaired_luis is not None:
            GlobalConfig.PRINT_UNPAIRED_LUIS_DEBUG_INFO = args.debug_unpaired_luis



# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from . import Utils

from .GlobalConfig import GlobalConfig, InputEndian
from .Context import Context, SymbolSpecialType, ContextSymbolBase, ContextSymbol, ContextOffsetSymbol, ContextRelocSymbol
from .FileSectionType import FileSectionType, FileSections_ListBasic, FileSections_ListAll
from .FileSplitFormat import FileSplitFormat, FileSplitEntry
from .ElementBase import ElementBase



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .. import common

from .Elf32Constants import Elf32SectionHeaderType
from .Elf32Header import Elf32Header
from .Elf32SectionHeaders import Elf32SectionHeaders
from .Elf32StringTable import Elf32StringTable
from .Elf32Syms import Elf32Syms
from .Elf32Rels import Elf32Rels


class Elf32File:
    def __init__(self, array_of_bytes: bytearray):
        self.header = Elf32Header.fromBytearray(array_of_bytes)
        # print(self.header)

        self.strtab: Elf32StringTable | None = None
        self.symtab: Elf32Syms | None = None

        self.progbits: dict[common.FileSectionType, bytearray] = dict()
        self.nobits: int | None = None

        self.rel: dict[common.FileSectionType, Elf32Rels] = dict()

        self.sectionHeaders = Elf32SectionHeaders(array_of_bytes, self.header.shoff, self.header.shnum)

        shstrtabSectionEntry = self.sectionHeaders.sections[self.header.shstrndx]
        self.shstrtab = Elf32StringTable(array_of_bytes, shstrtabSectionEntry.offset, shstrtabSectionEntry.size)

        for entry in self.sectionHeaders.sections:
            sectionEntryName = self.shstrtab[entry.name]
            # print(sectionEntryName, end="\t ")
            # print(entry)
            if entry.type == Elf32SectionHeaderType.NULL.value:
                continue
            elif entry.type == Elf32SectionHeaderType.PROGBITS.value:
                fileSecType = common.FileSectionType.fromStr(sectionEntryName)
                if fileSecType != common.FileSectionType.Invalid:
                    self.progbits[fileSecType] = array_of_bytes[entry.offset:entry.offset+entry.size]
                    common.Utils.printVerbose(sectionEntryName, "size: ", len(self.progbits[fileSecType]))
                    common.Utils.printVerbose()
                else:
                    common.Utils.eprint("Unknown PROGBITS found: ", sectionEntryName, entry)
            elif entry.type == Elf32SectionHeaderType.SYMTAB.value:
                if sectionEntryName == ".symtab":
                    self.symtab = Elf32Syms(array_of_bytes, entry.offset, entry.size)
                    common.Utils.printVerbose()
                    common.Utils.printVerbose("SYMTAB:")
                    for i, sym in enumerate(self.symtab.symbols):
                        common.Utils.printVerbose(i, sym)
                    common.Utils.printVerbose()
                else:
                    common.Utils.eprint("Unknown SYMTAB found: ", sectionEntryName, entry)
            elif entry.type == Elf32SectionHeaderType.STRTAB.value:
                if sectionEntryName == ".strtab":
                    self.strtab = Elf32StringTable(array_of_bytes, entry.offset, entry.size)
                    common.Utils.printVerbose()
                    common.Utils.printVerbose("STRTAB:")
                    for i, string in enumerate(self.strtab):
                        common.Utils.printVerbose(i, string)
                    common.Utils.printVerbose()
                elif sectionEntryName == ".shstrtab":
                    pass
                else:
                    common.Utils.eprint("Unknown STRTAB found: ", sectionEntryName, entry)
            # elif entry.type == Elf32SectionHeaderType.RELA.value:
            #     pass
            # elif entry.type == Elf32SectionHeaderType.HASH.value:
            #     pass
            # elif entry.type == Elf32SectionHeaderType.DYNAMIC.value:
            #     pass
            # elif entry.type == Elf32SectionHeaderType.NOTE.value:
            #     pass
            elif entry.type == Elf32SectionHeaderType.NOBITS.value:
                if sectionEntryName == ".bss":
                    self.nobits = entry.size
                    common.Utils.printVerbose(sectionEntryName, "size: ", self.nobits)
                    common.Utils.printVerbose()
                else:
                    common.Utils.eprint("Unknown NOBITS found: ", sectionEntryName, entry)
            elif entry.type == Elf32SectionHeaderType.REL.value:
                if sectionEntryName.startswith(".rel."):
                    fileSecType = common.FileSectionType.fromStr(sectionEntryName[4:])
                    if fileSecType != common.FileSectionType.Invalid:
                        self.rel[fileSecType] = Elf32Rels(array_of_bytes, entry.offset, entry.size)
                        common.Utils.printVerbose()
                        common.Utils.printVerbose(f"REL: ({sectionEntryName})")
                        for i, rel in enumerate(self.rel[fileSecType]):
                            common.Utils.printVerbose(i, rel, rel.rType, rel.rSym)
                        common.Utils.printVerbose()
                    else:
                        common.Utils.eprint("Unknown REL subsection found: ", sectionEntryName, entry)
                else:
                    common.Utils.eprint("Unknown REL found: ", sectionEntryName, entry)
            elif entry.type == Elf32SectionHeaderType.MIPS_GPTAB.value:
                # ?
                pass
            elif entry.type == Elf32SectionHeaderType.MIPS_DEBUG.value:
                # ?
                pass
            elif entry.type == Elf32SectionHeaderType.MIPS_REGINFO.value:
                # ?
                pass
            elif entry.type == Elf32SectionHeaderType.MIPS_OPTIONS.value:
                # ?
                pass
            elif entry.type == Elf32SectionHeaderType.MIPS_ABIFLAGS.value:
                # ?
                pass
            else:
                common.Utils.eprint("Unknown section header type found:", sectionEntryName, entry)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .MipsInstructionConfig import InstructionConfig, AbiNames
from .MipsConstants import InstructionId, InstructionVectorId, InstrType, InstrDescriptor, instructionDescriptorDict, InstructionsNotEmitedByIDO
from .MipsInstructionBase import InstructionBase

from .MipsInstructionNormal import InstructionNormal
from .MipsInstructionSpecial import InstructionSpecial
from .MipsInstructionRegimm import InstructionRegimm
from .MipsInstructionCoprocessor0 import InstructionCoprocessor0
from .MipsInstructionCoprocessor1 import InstructionCoprocessor1
from .MipsInstructionCoprocessor2 import InstructionCoprocessor2

from .MipsInstructionNormalRsp import InstructionNormalRsp
from .MipsInstructionSpecialRsp import InstructionSpecialRsp
from .MipsInstructionRegimmRsp import InstructionRegimmRsp
from .MipsInstructionCoprocessor0Rsp import InstructionCoprocessor0Rsp
from .MipsInstructionCoprocessor2Rsp import InstructionCoprocessor2Rsp

from .MipsInstructions import wordToInstruction, wordToInstructionRsp



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import ast
import argparse
import bisect
import dataclasses
import enum
import os

from . import Utils
from .GlobalConfig import GlobalConfig
from .FileSectionType import FileSectionType


class ContextSegment:
    def __init__(self, segmentName: str, segmentInputPath: str, segmentType: str, subsections):
        self.name: str = segmentName
        self.inputPath: str = segmentInputPath
        self.type: str = segmentType
        self.subsections: list = subsections

class SymbolSpecialType(enum.Enum):
    function        = enum.auto()
    branchlabel     = enum.auto()
    jumptable       = enum.auto()
    jumptablelabel  = enum.auto()
    fakefunction    = enum.auto()
    hardwarereg     = enum.auto()
    constant        = enum.auto()


    def toStr(self) -> str:
        return "@" + self.name

    @staticmethod
    def fromStr(symTypeStr: str|None) -> SymbolSpecialType|None:
        if symTypeStr == "@function":
            return SymbolSpecialType.function
        if symTypeStr == "@branchlabel":
            return SymbolSpecialType.branchlabel
        if symTypeStr == "@jumptable":
            return SymbolSpecialType.jumptable
        if symTypeStr == "@jumptablelabel":
            return SymbolSpecialType.jumptablelabel
        if symTypeStr == "@fakefunction":
            return SymbolSpecialType.fakefunction
        if symTypeStr == "@hardwarereg":
            return SymbolSpecialType.hardwarereg
        if symTypeStr == "@constant":
            return SymbolSpecialType.constant
        return None


@dataclasses.dataclass
class ContextSymbolBase:
    name: str
    size: int = 4
    type: SymbolSpecialType|str|None = None

    sectionType: FileSectionType = FileSectionType.Unknown

    isDefined: bool = False
    "This symbol exists in any of the analyzed sections"
    isUserDeclared: bool = False
    "Declared externally by the user, but it may have not been found yet"
    isAutogenerated: bool = False
    "This symbol was automatically generated by the disassembler"

    isMaybeString: bool = False

    referenceCounter: int = 0
    "How much this symbol is referenced by something else"

    address: int = 0
    "generic variable as a placeholder for vram or offset"


    def hasNoType(self) -> bool:
        return self.type is None or self.type == ""


    def isTrustableFunction(self, rsp: bool=False) -> bool:
        """Checks if the function symbol should be trusted based on the current disassembler settings.

        Note: This function doesn't check if the type of the symbol is SymbolSpecialType.function
        """
        if GlobalConfig.TRUST_USER_FUNCTIONS and self.isUserDeclared:
            return True

        if GlobalConfig.TRUST_JAL_FUNCTIONS and self.isAutogenerated:
            return True

        if rsp:
            return True

        return False


    def isByte(self) -> bool:
        if not GlobalConfig.USE_DOT_BYTE:
            return False
        return self.type in ("u8", "s8")

    def isShort(self) -> bool:
        if not GlobalConfig.USE_DOT_SHORT:
            return False
        return self.type in ("u16", "s16")


    def isString(self) -> bool:
        if self.type == "char" or self.type == "char*":
            return True
        elif self.hasNoType(): # no type information, let's try to guess
            if GlobalConfig.STRING_GUESSER and self.isMaybeString:
                return True
        return False

    def isFloat(self) -> bool:
        return self.type in ("f32", "Vec3f")

    def isDouble(self) -> bool:
        return self.type == "f64"

    def isJumpTable(self) -> bool:
        return self.type == SymbolSpecialType.jumptable

    def isMaybeConstVariable(self) -> bool:
        if self.isFloat():
            return False
        if self.isDouble():
            return False
        elif self.isJumpTable():
            return False
        elif self.isString():
            return False
        return True


    def isStatic(self) -> bool:
        return self.name.startswith(".")

    def isLateRodata(self) -> bool:
        # if self.referenceCounter > 1: return False # ?
        return self.isJumpTable() or self.isFloat() or self.isDouble()


    def setTypeIfUnset(self, varType: str) -> bool:
        if self.hasNoType():
            self.type = varType
            return True
        return False


    def getName(self) -> str:
        return self.name

    def getType(self) -> str:
        if self.type is None:
            return ""
        if isinstance(self.type, SymbolSpecialType):
            return self.type.toStr()
        return self.type

    def getSymbolPlusOffset(self, address: int) -> str:
        if self.address == address:
            return self.name
        if self.address > address:
            return f"{self.name} - 0x{self.address - address:X}"
        return f"{self.name} + 0x{address - self.address:X}"

    def getSymbolLabel(self) -> str:
        label = ""
        if self.isStatic():
            label += "/* static variable */" + GlobalConfig.LINE_ENDS
        if self.sectionType == FileSectionType.Text:
            label += GlobalConfig.ASM_TEXT_LABEL
        else:
            label += GlobalConfig.ASM_DATA_LABEL
        label += " " + self.getName()
        return label


    def toCsv(self) -> str:
        return f"0x{self.address:06X},{self.name},{self.getType()},0x{self.size:X},{self.sectionType.toStr()},{self.isDefined},{self.isUserDeclared},{self.isAutogenerated},{self.isMaybeString},{self.referenceCounter}"


class ContextSymbol(ContextSymbolBase):
    def __init__(self, vram: int, name: str, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.address = vram

    @property
    def vram(self) -> int:
        return self.address

class ContextOffsetSymbol(ContextSymbolBase):
    def __init__(self, offset: int, name: str, sectionType: FileSectionType, *args, **kwargs):
        super().__init__(name, *args, sectionType=sectionType, **kwargs)
        self.address = offset

    # Relative to the start of the section
    @property
    def offset(self) -> int:
        return self.address

    def getName(self) -> str:
        if self.isStatic():
            return self.name[1:]
        return self.name


class ContextRelocSymbol(ContextSymbolBase):
    relocSection: FileSectionType
    relocType: int = -1 # Same number as the .elf specification

    def __init__(self, offset: int, name: str, relocSection: FileSectionType, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.address = offset
        self.relocSection = relocSection

    # Relative to the start of the section
    @property
    def offset(self) -> int:
        return self.address

    def getNamePlusOffset(self, offset: int) -> str:
        if offset == 0:
            return self.name
        if offset < 0:
            return f"{self.name} - 0x{-offset:X}"
        return f"{self.name} + 0x{offset:X}"

    def toCsv(self) -> str:
        return super().toCsv() + f",{self.relocSection.toStr()},{self.relocType}"


class Context:
    N64DefaultBanned = {0x80000010, 0x80000020}

    N64LibultraSyms = {
        0x80000300: ("osTvType",       "u32", 0x4),
        0x80000304: ("osRomType",      "u32", 0x4),
        0x80000308: ("osRomBase",      "u32", 0x4),
        0x8000030C: ("osResetType",    "u32", 0x4),
        0x80000310: ("osCicId",        "u32", 0x4),
        0x80000314: ("osVersion",      "u32", 0x4),
        0x80000304: ("osRomType",      "u32", 0x4),
        0x80000318: ("osMemSize",      "u32", 0x4),
        0x8000031C: ("osAppNmiBuffer", "u8",  0x40),
    }

    N64HardwareRegs = {
        # Signal Processor Registers
        0xA4040000: "SP_MEM_ADDR_REG",
        0xA4040004: "SP_DRAM_ADDR_REG",
        0xA4040008: "SP_RD_LEN_REG",
        0xA404000C: "SP_WR_LEN_REG",
        0xA4040010: "SP_STATUS_REG",
        0xA4040014: "SP_DMA_FULL_REG",
        0xA4040018: "SP_DMA_BUSY_REG",
        0xA404001C: "SP_SEMAPHORE_REG",

        0xA4080000: "SP_PC",

        # Display Processor Command Registers / Rasterizer Interface
        0xA4100000: "DPC_START_REG",
        0xA4100004: "DPC_END_REG",
        0xA4100008: "DPC_CURRENT_REG",
        0xA410000C: "DPC_STATUS_REG",
        0xA4100010: "DPC_CLOCK_REG",
        0xA4100014: "DPC_BUFBUSY_REG",
        0xA4100018: "DPC_PIPEBUSY_REG",
        0xA410001C: "DPC_TMEM_REG",

        # Display Processor Span Registers
        0xA4200000: "DPS_TBIST_REG", # DPS_TBIST_REG / DP_TMEM_BIST
        0xA4200004: "DPS_TEST_MODE_REG",
        0xA4200008: "DPS_BUFTEST_ADDR_REG",
        0xA420000C: "DPS_BUFTEST_DATA_REG",

        # MIPS Interface Registers
        0xA4300000: "MI_MODE_REG", # MI_MODE_REG / MI_INIT_MODE_REG
        0xA4300004: "MI_VERSION_REG",
        0xA4300008: "MI_INTR_REG",
        0xA430000C: "MI_INTR_MASK_REG",

        # Video Interface Registers
        0xA4400000: "VI_STATUS_REG", # VI_STATUS_REG / VI_CONTROL_REG
        0xA4400004: "VI_DRAM_ADDR_REG", # VI_DRAM_ADDR_REG / VI_ORIGIN_REG
        0xA4400008: "VI_WIDTH_REG",
        0xA440000C: "VI_INTR_REG",
        0xA4400010: "VI_CURRENT_REG",
        0xA4400014: "VI_BURST_REG", # VI_BURST_REG / VI_TIMING_REG
        0xA4400018: "VI_V_SYNC_REG",
        0xA440001C: "VI_H_SYNC_REG",
        0xA4400020: "VI_LEAP_REG",
        0xA4400024: "VI_H_START_REG",
        0xA4400028: "VI_V_START_REG",
        0xA440002C: "VI_V_BURST_REG",
        0xA4400030: "VI_X_SCALE_REG",
        0xA4400034: "VI_Y_SCALE_REG",

        # Audio Interface Registers
        0xA4500000: "AI_DRAM_ADDR_REG",
        0xA4500004: "AI_LEN_REG",
        0xA4500008: "AI_CONTROL_REG",
        0xA450000C: "AI_STATUS_REG",
        0xA4500010: "AI_DACRATE_REG",
        0xA4500014: "AI_BITRATE_REG",

        # Peripheral/Parallel Interface Registers
        0xA4600000: "PI_DRAM_ADDR_REG",
        0xA4600004: "PI_CART_ADDR_REG",
        0xA4600005: "D_A4600005", # TODO: figure out its name
        0xA4600006: "D_A4600006", # TODO: figure out its name
        0xA4600007: "D_A4600007", # TODO: figure out its name
        0xA4600008: "PI_RD_LEN_REG",
        0xA460000C: "PI_WR_LEN_REG",
        0xA4600010: "PI_STATUS_REG",
        0xA4600014: "PI_BSD_DOM1_LAT_REG", # PI dom1 latency
        0xA4600018: "PI_BSD_DOM1_PWD_REG", # PI dom1 pulse width
        0xA460001C: "PI_BSD_DOM1_PGS_REG", # PI dom1 page size
        0xA4600020: "PI_BSD_DOM1_RLS_REG", # PI dom1 release
        0xA4600024: "PI_BSD_DOM2_LAT_REG", # PI dom2 latency
        0xA4600028: "PI_BSD_DOM2_LWD_REG", # PI dom2 pulse width
        0xA460002C: "PI_BSD_DOM2_PGS_REG", # PI dom2 page size
        0xA4600030: "PI_BSD_DOM2_RLS_REG", # PI dom2 release

        # RDRAM Interface Registers
        0xA4700000: "RI_MODE_REG",
        0xA4700004: "RI_CONFIG_REG",
        0xA4700008: "RI_CURRENT_LOAD_REG",
        0xA470000C: "RI_SELECT_REG",
        0xA4700010: "RI_REFRESH_REG",
        0xA4700014: "RI_LATENCY_REG",
        0xA4700018: "RI_RERROR_REG",
        0xA470001C: "RI_WERROR_REG",

        # Serial Interface Registers
        0xA4800000: "SI_DRAM_ADDR_REG",
        0xA4800004: "SI_PIF_ADDR_RD64B_REG",
        0xA4800008: "D_A4800008", # reserved
        0xA480000C: "D_A480000C", # reserved
        0xA4800010: "SI_PIF_ADDR_WR64B_REG",
        0xA4800014: "D_A4800014", # reserved
        0xA4800018: "SI_STATUS_REG",
    }
    "N64 OS hardware registers"

    def __init__(self):
        self.segments: dict[str, ContextSegment] = dict()

        self.funcsInFiles: dict[str, list[int]] = dict()

        self.functions: dict[int, ContextSymbol] = dict()

        self.labels: dict[int, ContextSymbol] = dict()

        self.symbols: dict[int, ContextSymbol] = dict()
        self.symbolsVramSorted: list[int] = list()

        # Functions jumped into Using J instead of JAL
        self.fakeFunctions: dict[int, ContextSymbol] = dict()

        self.constants: dict[int, ContextSymbol] = dict()

        self.newPointersInData: set[int] = set()

        self.loPatches: dict[int, int] = dict()
        "key: address of %lo, value: symbol's vram to use instead"

        self.dataSymbolsWithReferencesWithAddends: set[int] = set()
        "Contains the address of data symbols which are allowed to have references to other symbols with addends"

        self.dataReferencingConstants: set[int] = set()
        "Set of addresses of data symbols which are allowed to reference named constants"

        # Stuff that looks like pointers, but the disassembler shouldn't count it as a pointer
        self.bannedSymbols: set[int] = set()

        # First key is the section type, sub key is offset relative to the start of that section
        self.offsetSymbols: dict[FileSectionType, dict[int, ContextOffsetSymbol]] = {
            FileSectionType.Text: dict(),
            FileSectionType.Data: dict(),
            FileSectionType.Rodata: dict(),
            FileSectionType.Bss: dict(),
        }

        self.relocSymbols: dict[FileSectionType, dict[int, ContextRelocSymbol]] = {
            FileSectionType.Text: dict(),
            FileSectionType.Data: dict(),
            FileSectionType.Rodata: dict(),
            FileSectionType.Bss: dict(),
        }

        # Where the jump table is
        self.offsetJumpTables: dict[int, ContextOffsetSymbol] = dict()
        # The addresses every jump table has
        self.offsetJumpTablesLabels: dict[int, ContextOffsetSymbol] = dict()


    def getAnySymbol(self, vramAddress: int) -> ContextSymbol|None:
        "Searches for any known function, label and symbol"
        if vramAddress == 0:
            return None

        if vramAddress in self.functions:
            return self.functions[vramAddress]

        if vramAddress in self.labels:
            return self.labels[vramAddress]

        if vramAddress in self.symbols:
            return self.symbols[vramAddress]

        if vramAddress in self.fakeFunctions:
            return self.fakeFunctions[vramAddress]

        return None

    def getGenericLabel(self, vramAddress: int) -> ContextSymbol|None:
        "Searches for any .text label (function, branch label or jtbl label)"
        if vramAddress == 0:
            return None

        if vramAddress in self.functions:
            return self.functions[vramAddress]

        if vramAddress in self.labels:
            return self.labels[vramAddress]

        if vramAddress in self.fakeFunctions:
            return self.fakeFunctions[vramAddress]

        return None

    def getGenericSymbol(self, vramAddress: int, tryPlusOffset: bool = True, checkUpperLimit: bool = True) -> ContextSymbol|None:
        "Searches a function or a symbol, or optionally for a symbol with an addend if `tryPlusOffset` is True"
        if vramAddress == 0:
            return None

        if vramAddress in self.functions:
            return self.functions[vramAddress]

        if vramAddress in self.symbols:
            return self.symbols[vramAddress]

        if vramAddress in self.fakeFunctions:
            return self.fakeFunctions[vramAddress]

        if GlobalConfig.PRODUCE_SYMBOLS_PLUS_OFFSET and tryPlusOffset:
            vramIndex = bisect.bisect(self.symbolsVramSorted, vramAddress)
            if vramIndex != len(self.symbolsVramSorted):
                symVram = self.symbolsVramSorted[vramIndex-1]
                contextSym = self.symbols[symVram]

                if vramAddress > symVram:
                    if checkUpperLimit and vramAddress >= symVram + contextSym.size:
                        return None
                    return contextSym
        return None

    def getSymbol(self, vramAddress: int, tryPlusOffset: bool = True, checkUpperLimit: bool = True) -> ContextSymbol|None:
        "Searches symbol or a symbol with an addend if `tryPlusOffset` is True"
        if vramAddress == 0:
            return None

        if vramAddress in self.symbols:
            return self.symbols[vramAddress]

        if GlobalConfig.PRODUCE_SYMBOLS_PLUS_OFFSET and tryPlusOffset:
            vramIndex = bisect.bisect(self.symbolsVramSorted, vramAddress)
            if vramIndex != len(self.symbolsVramSorted):
                symVram = self.symbolsVramSorted[vramIndex-1]
                contextSym = self.symbols[symVram]

                symbolSize = contextSym.size
                if vramAddress > symVram:
                    if checkUpperLimit:
                        if vramAddress >= symVram + symbolSize:
                            return None
                    return contextSym
        return None

    def getFunction(self, vramAddress: int) -> ContextSymbol|None:
        if vramAddress == 0:
            return None

        if vramAddress in self.functions:
            return self.functions[vramAddress]

        return None

    def getConstant(self, constantValue: int) -> ContextSymbol|None:
        if constantValue in self.constants:
            return self.constants[constantValue]

        return None

    def getLoPatch(self, loInstrVram: int|None) -> int|None:
        if loInstrVram is None:
            return None
        return self.loPatches.get(loInstrVram, None)


    def getOffsetSymbol(self, offset: int, sectionType: FileSectionType) -> ContextOffsetSymbol|None:
        if sectionType in self.offsetSymbols:
            symbolsInSection = self.offsetSymbols[sectionType]
            if offset in symbolsInSection:
                return symbolsInSection[offset]
        return None

    def getOffsetGenericSymbol(self, offset: int, sectionType: FileSectionType) -> ContextOffsetSymbol|None:
        if offset in self.offsetJumpTables:
            return self.offsetJumpTables[offset]

        if sectionType in self.offsetSymbols:
            symbolsInSection = self.offsetSymbols[sectionType]
            if offset in symbolsInSection:
                return symbolsInSection[offset]

        return None

    def getRelocSymbol(self, offset: int, sectionType: FileSectionType) -> ContextRelocSymbol|None:
        if sectionType in self.relocSymbols:
            relocsInSection = self.relocSymbols[sectionType]
            if offset in relocsInSection:
                return relocsInSection[offset]
        return None

    def getOffsetGenericLabel(self, offset: int, sectionType: FileSectionType) -> ContextOffsetSymbol|None:
        if offset in self.offsetJumpTablesLabels:
            return self.offsetJumpTablesLabels[offset]
        return None


    def addSymbol(self, vramAddress: int, name: str|None, sectionType: FileSectionType=FileSectionType.Unknown, isAutogenerated: bool=False) -> ContextSymbol:
        if name is None:
            name = f"D_{vramAddress:06X}"
            if GlobalConfig.AUTOGENERATED_NAMES_BASED_ON_SECTION_TYPE:
                if sectionType == FileSectionType.Rodata:
                    name = f"R_{vramAddress:06X}"
                elif sectionType == FileSectionType.Bss:
                    name = f"B_{vramAddress:06X}"
        if vramAddress not in self.symbols:
            contextSymbol = ContextSymbol(vramAddress, name)
            contextSymbol.isAutogenerated = isAutogenerated
            contextSymbol.sectionType = sectionType
            self.symbols[vramAddress] = contextSymbol
            bisect.insort(self.symbolsVramSorted, vramAddress)
        else:
            contextSymbol = self.symbols[vramAddress]
        if contextSymbol.isAutogenerated:
            contextSymbol.name = name
        return contextSymbol

    def addFunction(self, vramAddress: int, name: str|None=None, isAutogenerated: bool=False) -> ContextSymbol:
        if vramAddress not in self.functions:
            if name is None:
                name = f"func_{vramAddress:08X}"
            contextSymbol = ContextSymbol(vramAddress, name)
            contextSymbol.type = SymbolSpecialType.function
            contextSymbol.isAutogenerated = isAutogenerated
            self.functions[vramAddress] = contextSymbol
        else:
            contextSymbol = self.functions[vramAddress]
        contextSymbol.sectionType = FileSectionType.Text

        if vramAddress in self.fakeFunctions:
            del self.fakeFunctions[vramAddress]

        return contextSymbol

    def addBranchLabel(self, vramAddress: int, name: str|None=None, isAutogenerated: bool=False) -> ContextSymbol:
        if name is None:
            name = f".L{vramAddress:08X}"
        if vramAddress not in self.labels:
            contextSymbol = ContextSymbol(vramAddress, name)
            contextSymbol.isAutogenerated = isAutogenerated
            self.labels[vramAddress] = contextSymbol
        else:
            contextSymbol = self.labels[vramAddress]
        contextSymbol.type = SymbolSpecialType.branchlabel
        if contextSymbol.isAutogenerated:
            contextSymbol.name = name
        return contextSymbol

    def addJumpTable(self, vramAddress: int, name: str|None=None, isAutogenerated: bool=False) -> ContextSymbol:
        if name is None:
            name = f"jtbl_{vramAddress:08X}"
        contextSymbol = self.addSymbol(vramAddress, name, isAutogenerated=isAutogenerated)
        contextSymbol.type = SymbolSpecialType.jumptable
        if contextSymbol.isAutogenerated:
            contextSymbol.name = name
        return contextSymbol

    def addJumpTableLabel(self, vramAddress: int, name: str|None=None, isAutogenerated: bool=False) -> ContextSymbol:
        if name is None:
            name = f"L{vramAddress:08X}"
        if vramAddress not in self.labels:
            contextSymbol = ContextSymbol(vramAddress, name)
            contextSymbol.isAutogenerated = isAutogenerated
            self.labels[vramAddress] = contextSymbol
        else:
            contextSymbol = self.labels[vramAddress]
        contextSymbol.type = SymbolSpecialType.jumptablelabel
        if contextSymbol.isAutogenerated:
            contextSymbol.name = name
        return contextSymbol

    def addFakeFunction(self, vramAddress: int, name: str, isAutogenerated: bool=False) -> ContextSymbol:
        if vramAddress not in self.fakeFunctions:
            contextSymbol = ContextSymbol(vramAddress, name)
            contextSymbol.type = SymbolSpecialType.fakefunction
            contextSymbol.isAutogenerated = isAutogenerated
            self.fakeFunctions[vramAddress] = contextSymbol
            return contextSymbol
        return self.fakeFunctions[vramAddress]

    def addConstant(self, constantValue: int, name: str) -> ContextSymbol:
        if constantValue not in self.constants:
            contextSymbol = ContextSymbol(constantValue, name)
            contextSymbol.type = SymbolSpecialType.constant
            self.constants[constantValue] = contextSymbol
            return contextSymbol
        return self.constants[constantValue]

    def addOffsetJumpTable(self, offset: int, sectionType: FileSectionType) -> ContextOffsetSymbol:
        if offset not in self.offsetJumpTables:
            contextOffsetSym = ContextOffsetSymbol(offset, f"jtbl_{offset:06X}", sectionType)
            contextOffsetSym.type = SymbolSpecialType.jumptable
            self.offsetJumpTables[offset] = contextOffsetSym
            return contextOffsetSym
        return self.offsetJumpTables[offset]

    def addOffsetJumpTableLabel(self, offset: int, name: str, sectionType: FileSectionType) -> ContextOffsetSymbol:
        if offset not in self.offsetJumpTablesLabels:
            contextOffsetSym = ContextOffsetSymbol(offset, name, sectionType)
            contextOffsetSym.type = SymbolSpecialType.jumptablelabel
            self.offsetJumpTablesLabels[offset] = contextOffsetSym
            return contextOffsetSym
        return self.offsetJumpTablesLabels[offset]


    def fillDefaultBannedSymbols(self):
        self.bannedSymbols |= self.N64DefaultBanned

    def fillLibultraSymbols(self):
        for vram, (name, type, size) in self.N64LibultraSyms.items():
            contextSym = self.addSymbol(vram, name)
            contextSym.type = type
            contextSym.size = size
            contextSym.isDefined = True
            contextSym.isUserDeclared = True

    def fillHardwareRegs(self, useRealNames: bool=False):
        for vram, name in self.N64HardwareRegs.items():
            nameToUse = None
            if useRealNames:
                nameToUse = name
            contextSym = self.addSymbol(vram, nameToUse)
            contextSym.type = SymbolSpecialType.hardwarereg
            contextSym.size = 4
            contextSym.isDefined = True
            contextSym.isUserDeclared = True


    def readMMAddressMaps(self, filesPath: str, functionsPath: str, variablesPath: str):
        with open(filesPath) as infile:
            files_spec = ast.literal_eval(infile.read())

        for segmentName, segmentInputPath, segmentType, subsections, subfiles  in files_spec:
            self.segments[segmentName] = ContextSegment(segmentName, segmentInputPath, segmentType, subsections)
            for vram, subname in subfiles.items():
                if subname == "":
                    subname = f"{segmentName}_{vram:08X}"

        with open(functionsPath) as infile:
            functions_ast = ast.literal_eval(infile.read())

        for vram, funcData in functions_ast.items():
            funcName = funcData[0]
            self.addFunction(vram, funcName)
            self.functions[vram].isUserDeclared = True

        with open(variablesPath) as infile:
            variables_ast = ast.literal_eval(infile.read())

        for vram, varData in variables_ast.items():
            varName, varType, varArrayInfo, varSize = varData
            if varType == "":
                varType = None

            contextSym = self.addSymbol(vram, varName)
            contextSym.type = varType
            contextSym.size = varSize
            contextSym.isUserDeclared = True

    def readVariablesCsv(self, filepath: str):
        if not os.path.exists(filepath):
            return

        variables_file = Utils.readCsv(filepath)
        for row in variables_file:
            if len(row) == 0:
                continue

            varType: SymbolSpecialType|str|None
            vramStr, varName, varType, varSizeStr = row

            if vramStr == "-":
                continue

            vram = int(vramStr, 16)
            varSize = int(varSizeStr, 16)
            if varType == "":
                varType = None

            specialType = SymbolSpecialType.fromStr(varType)
            if specialType is not None:
                varType = specialType
                if specialType == SymbolSpecialType.function:
                    contextSym = self.addFunction(vram, varName)
                elif specialType == SymbolSpecialType.branchlabel:
                    contextSym = self.addBranchLabel(vram, varName)
                elif specialType == SymbolSpecialType.jumptable:
                    contextSym = self.addJumpTable(vram, varName)
                elif specialType == SymbolSpecialType.jumptablelabel:
                    contextSym = self.addJumpTableLabel(vram, varName)
                elif specialType == SymbolSpecialType.fakefunction:
                    contextSym = self.addFakeFunction(vram, varName)
                elif specialType == SymbolSpecialType.hardwarereg:
                    contextSym = self.addSymbol(vram, varName)
                else:
                    contextSym = self.addSymbol(vram, varName)
            else:
                contextSym = self.addSymbol(vram, varName)

            contextSym.type = varType
            contextSym.size = varSize
            contextSym.isUserDeclared = True

    def readFunctionsCsv(self, filepath: str):
        if not os.path.exists(filepath):
            return

        functions_file = Utils.readCsv(filepath)
        for row in functions_file:
            if len(row) == 0:
                continue

            vramStr, funcName = row

            if vramStr == "-":
                continue

            vram = int(vramStr, 16)
            contextSym = self.addFunction(vram, funcName)
            contextSym.isUserDeclared = True

    def readConstantsCsv(self, filepath: str):
        if not os.path.exists(filepath):
            return

        constants_file = Utils.readCsv(filepath)
        for row in constants_file:
            if len(row) == 0:
                continue

            constantValueStr, constantName = row

            if constantValueStr == "-":
                continue

            constantValue = int(constantValueStr, 16)
            contextSym = self.addConstant(constantValue, constantName)
            contextSym.isUserDeclared = True


    def saveContextToFile(self, filepath: str):
        with open(filepath, "w") as f:
            for address, symbol in self.functions.items():
                f.write(f"function,{symbol.toCsv()}\n")

            for address, symbol in self.labels.items():
                f.write(f"label,{symbol.toCsv()}\n")

            for address, symbol in self.symbols.items():
                f.write(f"symbol,{symbol.toCsv()}\n")

            for address, symbol in self.fakeFunctions.items():
                f.write(f"fake_function,{symbol.toCsv()}\n")

            for address, constant in self.constants.items():
                f.write(f"constants,{constant.toCsv()}\n")

            for address in self.newPointersInData:
                f.write(f"new_pointer_in_data,0x{address:08X}\n")

            for address in self.bannedSymbols:
                f.write(f"banned_symbol,0x{address:08X}\n")


    @staticmethod
    def addParametersToArgParse(parser: argparse.ArgumentParser):
        contextParser = parser.add_argument_group("Context configuration")

        contextParser.add_argument("--save-context", help="Saves the context to a file", metavar="FILENAME")


        csvConfig = parser.add_argument_group("Context .csv input files")

        csvConfig.add_argument("--functions", help="Path to a functions csv", action="append")
        csvConfig.add_argument("--variables", help="Path to a variables csv", action="append")
        csvConfig.add_argument("--constants", help="Path to a constants csv", action="append")


        symbolsConfig = parser.add_argument_group("Context default symbols configuration")

        symbolsConfig.add_argument("--default-banned", help="Toggles filling the list of default banned symbols. Defaults to True", action=argparse.BooleanOptionalAction)
        symbolsConfig.add_argument("--libultra-syms", help="Toggles using the built-in libultra symbols. Defaults to True", action=argparse.BooleanOptionalAction)
        symbolsConfig.add_argument("--hardware-regs", help="Toggles using the built-in hardware registers symbols. Defaults to True", action=argparse.BooleanOptionalAction)
        symbolsConfig.add_argument("--named-hardware-regs", help="Use actual names for the hardware registers", action=argparse.BooleanOptionalAction)


    def parseArgs(self, args: argparse.Namespace):
        if args.default_banned != False:
            self.fillDefaultBannedSymbols()
        if args.libultra_syms != False:
            self.fillLibultraSymbols()
        if not args.hardware_regs != False:
            self.fillHardwareRegs(args.named_hardware_regs)

        if args.functions is not None:
            for funcsPath in args.functions:
                self.readFunctionsCsv(funcsPath)
        if args.variables is not None:
            for varsPath in args.variables:
                self.readVariablesCsv(varsPath)
        if args.constants is not None:
            for constantsPath in args.constants:
                self.readConstantsCsv(constantsPath)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Callable

from ... import common

from .MipsInstructionConfig import InstructionConfig, AbiNames
from .MipsConstants import InstructionId, InstructionVectorId, InstrType, InstrDescriptor, instructionDescriptorDict


class InstructionBase:
    GprO32RegisterNames = {
        0:  "$zero",
        1:  "$at",
        2:  "$v0",
        3:  "$v1",
        4:  "$a0",
        5:  "$a1",
        6:  "$a2",
        7:  "$a3",
        8:  "$t0",
        9:  "$t1",
        10: "$t2",
        11: "$t3",
        12: "$t4",
        13: "$t5",
        14: "$t6",
        15: "$t7",
        16: "$s0",
        17: "$s1",
        18: "$s2",
        19: "$s3",
        20: "$s4",
        21: "$s5",
        22: "$s6",
        23: "$s7",
        24: "$t8",
        25: "$t9",
        26: "$k0",
        27: "$k1",
        28: "$gp",
        29: "$sp",
        30: "$fp",
        31: "$ra",
    }
    GprN32RegisterNames = {
        0:  "$zero",
        1:  "$at",
        2:  "$v0",
        3:  "$v1",
        4:  "$a0",
        5:  "$a1",
        6:  "$a2",
        7:  "$a3",
        8:  "$a4",
        9:  "$a5",
        10: "$a6",
        11: "$a7",
        12: "$t0",
        13: "$t1",
        14: "$t2",
        15: "$t3",
        16: "$s0",
        17: "$s1",
        18: "$s2",
        19: "$s3",
        20: "$s4",
        21: "$s5",
        22: "$s6",
        23: "$s7",
        24: "$t8",
        25: "$t9",
        26: "$k0",
        27: "$k1",
        28: "$gp",
        29: "$sp",
        30: "$fp",
        31: "$ra",
    }

    Cop0RegisterNames = {
        0:  "Index",
        1:  "Random",
        2:  "EntryLo0",
        3:  "EntryLo1",
        4:  "Context",
        5:  "PageMask",
        6:  "Wired",
        7:  "Reserved07",
        8:  "BadVaddr",
        9:  "Count",
        10: "EntryHi",
        11: "Compare",
        12: "Status",
        13: "Cause",
        14: "EPC",
        15: "PRevID",
        16: "Config",
        17: "LLAddr",
        18: "WatchLo",
        19: "WatchHi",
        20: "XContext",
        21: "Reserved21",
        22: "Reserved22",
        23: "Reserved23",
        24: "Reserved24",
        25: "Reserved25",
        26: "PErr",
        27: "CacheErr",
        28: "TagLo",
        29: "TagHi",
        30: "ErrorEPC",
        31: "Reserved31",
    }

    # Float registers
    Cop1NumericRegisterNames = {
        0:  "$f0",
        1:  "$f1",
        2:  "$f2",
        3:  "$f3",
        4:  "$f4",
        5:  "$f5",
        6:  "$f6",
        7:  "$f7",
        8:  "$f8",
        9:  "$f9",
        10: "$f10",
        11: "$f11",
        12: "$f12",
        13: "$f13",
        14: "$f14",
        15: "$f15",
        16: "$f16",
        17: "$f17",
        18: "$f18",
        19: "$f19",
        20: "$f20",
        21: "$f21",
        22: "$f22",
        23: "$f23",
        24: "$f24",
        25: "$f25",
        26: "$f26",
        27: "$f27",
        28: "$f28",
        29: "$f29",
        30: "$f30",
        31: "FpcCsr",
    }
    Cop1O32RegisterNames = {
        0:  "$fv0",
        1:  "$fv0f",
        2:  "$fv1",
        3:  "$fv1f",
        4:  "$ft0",
        5:  "$ft0f",
        6:  "$ft1",
        7:  "$ft1f",
        8:  "$ft2",
        9:  "$ft2f",
        10: "$ft3",
        11: "$ft3f",
        12: "$fa0",
        13: "$fa0f",
        14: "$fa1",
        15: "$fa1f",
        16: "$ft4",
        17: "$ft4f",
        18: "$ft5",
        19: "$ft5f",
        20: "$fs0",
        21: "$fs0f",
        22: "$fs1",
        23: "$fs1f",
        24: "$fs2",
        25: "$fs2f",
        26: "$fs3",
        27: "$fs3f",
        28: "$fs4",
        29: "$fs4f",
        30: "$fs5",
        31: "$fs5f",
    }
    Cop1N32RegisterNames = {
        0:  "$fv0",
        1:  "$ft14",
        2:  "$fv1",
        3:  "$ft15",
        4:  "$ft0",
        5:  "$ft1",
        6:  "$ft2",
        7:  "$ft3",
        8:  "$ft4",
        9:  "$ft5",
        10: "$ft6",
        11: "$ft7",
        12: "$fa0",
        13: "$fa1",
        14: "$fa2",
        15: "$fa3",
        16: "$fa4",
        17: "$fa5",
        18: "$fa6",
        19: "$fa7",
        20: "$fs0",
        21: "$ft8",
        22: "$fs1",
        23: "$ft9",
        24: "$fs2",
        25: "$ft10",
        26: "$fs3",
        27: "$ft11",
        28: "$fs4",
        29: "$ft12",
        30: "$fs5",
        31: "$ft13",
    }
    Cop1N64RegisterNames = {
        0:  "$fv0",
        1:  "$ft12",
        2:  "$fv1",
        3:  "$ft13",
        4:  "$ft0",
        5:  "$ft1",
        6:  "$ft2",
        7:  "$ft3",
        8:  "$ft4",
        9:  "$ft5",
        10: "$ft6",
        11: "$ft7",
        12: "$fa0",
        13: "$fa1",
        14: "$fa2",
        15: "$fa3",
        16: "$fa4",
        17: "$fa5",
        18: "$fa6",
        19: "$fa7",
        20: "$ft8",
        21: "$ft9",
        22: "$ft10",
        23: "$ft11",
        24: "$fs0",
        25: "$fs1",
        26: "$fs2",
        27: "$fs3",
        28: "$fs4",
        29: "$fs5",
        30: "$fs6",
        31: "$fs7",
    }

    GprRspRegisterNames = {
        0:  "$zero",
        1:  "$1",
        2:  "$2",
        3:  "$3",
        4:  "$4",
        5:  "$5",
        6:  "$6",
        7:  "$7",
        8:  "$8",
        9:  "$9",
        10: "$10",
        11: "$11",
        12: "$12",
        13: "$13",
        14: "$14",
        15: "$15",
        16: "$16",
        17: "$17",
        18: "$18",
        19: "$19",
        20: "$20",
        21: "$21",
        22: "$22",
        23: "$23",
        24: "$24",
        25: "$25",
        26: "$26",
        27: "$27",
        28: "$28",
        29: "$29",
        30: "$30",
        31: "$31",
    }

    Cop0RspRegisterNames = {
        0:  "SP_MEM_ADDR",
        1:  "SP_DRAM_ADDR",
        2:  "SP_RD_LEN",
        3:  "SP_WR_LEN",
        4:  "SP_STATUS",
        5:  "SP_DMA_FULL",
        6:  "SP_DMA_BUSY",
        7:  "SP_SEMAPHORE",
        8:  "DPC_START",
        9:  "DPC_END",
        10: "DPC_CURRENT",
        11: "DPC_STATUS",
        12: "DPC_CLOCK",
        13: "DPC_BUFBUSY",
        14: "DPC_PIPEBUSY",
        15: "DPC_TMEM",
    }

    VectorRspRegisterNames = {
        0:  "$v0",
        1:  "$v1",
        2:  "$v2",
        3:  "$v3",
        4:  "$v4",
        5:  "$v5",
        6:  "$v6",
        7:  "$v7",
        8:  "$v8",
        9:  "$v9",
        10: "$v10",
        11: "$v11",
        12: "$v12",
        13: "$v13",
        14: "$v14",
        15: "$v15",
        16: "$v16",
        17: "$v17",
        18: "$v18",
        19: "$v19",
        20: "$v20",
        21: "$v21",
        22: "$v22",
        23: "$v23",
        24: "$v24",
        25: "$v25",
        26: "$v26",
        27: "$v27",
        28: "$v28",
        29: "$v29",
        30: "$v30",
        31: "$v31",
    }

    instrArgumentsCallbacks: dict[str, Callable[[InstructionBase, str|None], str]] = {
        "rs":        lambda instr, immOverride: instr.getRegisterName(instr.rs),
        "rt":        lambda instr, immOverride: instr.getRegisterName(instr.rt),
        "rd":        lambda instr, immOverride: instr.getRegisterName(instr.rd),
        "sa":        lambda instr, immOverride: str(instr.sa),
        "ft":        lambda instr, immOverride: instr.getFloatRegisterName(instr.ft),
        "fs":        lambda instr, immOverride: instr.getFloatRegisterName(instr.fs),
        "fd":        lambda instr, immOverride: instr.getFloatRegisterName(instr.fd),
        "IMM":       lambda instr, immOverride: instr.processImmediate(immOverride),
        "LABEL":     lambda instr, immOverride: immOverride if immOverride is not None else f"func_{instr.getInstrIndexAsVram():06X}",
        "cop2t":     lambda instr, immOverride: instr.getCop2RegisterName(instr.rt),
        "cop0d":     lambda instr, immOverride: instr.getCop0RegisterName(instr.rd),
        "code":      lambda instr, immOverride: instr.processCodeParameter(),
        "op":        lambda instr, immOverride: f"0x{instr.rt:02X}",
        "IMM(base)": lambda instr, immOverride: f"{instr.processImmediate(immOverride)}({instr.getRegisterName(instr.baseRegister)})",
    }
    """Dictionary of callbacks to process the operands of an instruction.

    The keys should match the ones used in InstrDescriptor#operands
    """

    def __init__(self, instr: int):
        self.opcode = (instr >> 26) & 0x3F
        self.rs = (instr >> 21) & 0x1F # rs
        self.rt = (instr >> 16) & 0x1F # usually the destiny of the operation
        self.rd = (instr >> 11) & 0x1F # destination register in R-Type instructions
        self.sa = (instr >>  6) & 0x1F
        self.function = (instr >> 0) & 0x3F

        self.uniqueId: InstructionId|InstructionVectorId = InstructionId.INVALID
        self.descriptor: InstrDescriptor = instructionDescriptorDict[self.uniqueId]

        self.extraLjustWidthOpcode = 0

        self.vram: int|None = None
        self._handwrittenCategory: bool = False
        self.inHandwrittenFunction: bool = False

    @property
    def instr(self) -> int:
        return (self.opcode << 26) | (self.rs << 21) | (self.rt << 16) | (self.rd << 11) | (self.sa << 6) | (self.function)

    @property
    def immediate(self) -> int:
        return (self.rd << 11) | (self.sa << 6) | (self.function)
    @property
    def instr_index(self) -> int:
        return (self.rs << 21) | (self.rt << 16) | (self.immediate)
    @property
    def baseRegister(self) -> int:
        return self.rs

    @property
    def fmt(self) -> int:
        return self.rs

    @property
    def ft(self) -> int:
        return self.rt
    @property
    def fs(self) -> int:
        return self.rd
    @property
    def fd(self) -> int:
        return self.sa

    @property
    def nd(self) -> int:
        return (self.rt >> 0) & 0x01
    @property
    def tf(self) -> int:
        return (self.rt >> 1) & 0x01
    @property
    def fc(self) -> int:
        return (self.function >> 4) & 0x03
    @property
    def cond(self) -> int:
        return (self.function >> 0) & 0x0F

    # vector registers
    @property
    def vd(self) -> int:
        return self.sa
    @property
    def vs(self) -> int:
        return self.rd
    @property
    def vt(self) -> int:
        return self.rt
    @property
    def elementHigh(self) -> int:
        return self.rs & 0xF
    @property
    def elementLow(self) -> int:
        return (self.sa >> 1) & 0xF
    @property
    def offsetVector(self) -> int:
        return self.immediate & 0x7F

    def getInstrIndexAsVram(self) -> int:
        vram = self.instr_index << 2
        if self.vram is None:
            vram |= 0x80000000
        else:
            # Jumps are PC-region branches. The upper bits are filled with the address in the delay slot
            vram |= (self.vram+4) & 0xFF000000
        return vram


    def __getitem__(self, key):
        if key < 0 or key > 31:
            raise IndexError()
        return (self.instr >> key) & 0x1


    def processUniqueId(self):
        if self.uniqueId in instructionDescriptorDict:
            self.descriptor = instructionDescriptorDict[self.uniqueId]

    def isImplemented(self) -> bool:
        if self.uniqueId == InstructionId.INVALID:
            return False
        if self.uniqueId == InstructionVectorId.INVALID:
            return False
        return True

    def isFloatInstruction(self) -> bool:
        return self.descriptor.isFloat

    def isDoubleFloatInstruction(self) -> bool:
        return self.descriptor.isDouble


    def isBranch(self) -> bool:
        return self.descriptor.isBranch
    def isBranchLikely(self) -> bool:
        return self.descriptor.isBranchLikely
    def isJump(self) -> bool:
        return self.descriptor.isJump
    def isTrap(self) -> bool:
        return self.descriptor.isTrap

    def isJType(self) -> bool:
        return self.descriptor.instrType == InstrType.typeJ

    def isIType(self) -> bool:
        return self.descriptor.instrType == InstrType.typeI

    def isRType(self) -> bool:
        return self.descriptor.instrType == InstrType.typeR


    def sameOpcode(self, other: InstructionBase) -> bool:
        if not self.isImplemented():
            return False
        if not other.isImplemented():
            return False
        return self.uniqueId == other.uniqueId

    def sameBaseRegister(self, other: InstructionBase):
        return self.baseRegister == other.baseRegister

    def sameOpcodeButDifferentArguments(self, other: InstructionBase) -> bool:
        if not self.sameOpcode(other):
            return False
        return self.instr != other.instr


    def modifiesRt(self) -> bool:
        return self.descriptor.modifiesRt
    def modifiesRd(self) -> bool:
        return self.descriptor.modifiesRd


    def blankOut(self):
        self.rs = 0
        self.rt = 0
        self.rd = 0
        self.sa = 0
        self.function = 0


    def getOpcodeName(self) -> str:
        if self.isImplemented():
            return self.uniqueId.name
        return f"(0x{self.opcode:02X})"


    def getRegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            return f"${register}"
        if InstructionConfig.GPR_ABI_NAMES == AbiNames.o32:
            return self.GprO32RegisterNames.get(register, f"${register}")
        if InstructionConfig.GPR_ABI_NAMES == AbiNames.numeric:
            return f"${register}"
        # AbiNames.n32 or AbiNames.n64
        return self.GprN32RegisterNames.get(register, f"${register}")

    def getFloatRegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            if register == 31:
                return f"${register}"
            return f"$f{register}"
        if InstructionConfig.FPR_ABI_NAMES == AbiNames.numeric:
            if register == 31 and not InstructionConfig.USE_FPCCSR:
                return "$31"
            return self.Cop1NumericRegisterNames.get(register, f"${register}")
        if InstructionConfig.FPR_ABI_NAMES == AbiNames.o32:
            return self.Cop1O32RegisterNames.get(register, f"${register}")
        if InstructionConfig.FPR_ABI_NAMES == AbiNames.n32:
            return self.Cop1N32RegisterNames.get(register, f"${register}")
        # AbiNames.n64
        return self.Cop1N64RegisterNames.get(register, f"${register}")

    def getCop0RegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            return f"${register}"
        if InstructionConfig.VR4300_COP0_NAMED_REGISTERS:
            return self.Cop0RegisterNames.get(register, f"${register}")
        return f"${register}"

    def getCop2RegisterName(self, register: int) -> str:
        return f"${register}"

    def getGprRspRegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            return f"${register}"
        return self.GprRspRegisterNames.get(register, f"${register}")

    def getCop0RspRegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            return f"${register}"
        if InstructionConfig.VR4300_RSP_COP0_NAMED_REGISTERS:
            return self.Cop0RspRegisterNames.get(register, f"${register}")
        return f"${register}"

    def getVectorRspRegisterName(self, register: int) -> str:
        if not InstructionConfig.NAMED_REGISTERS:
            return f"${register}"
        return self.VectorRspRegisterNames.get(register, f"${register}")


    def processVectorElement(self, element: int) -> int:
        if (element & 0x8) == 0x8:
            return element & 7
        if (element & 0xC) == 0x4:
            return element & 4
        if (element & 0xE) == 0x2:
            return element & 2
        return element

    def getBranchOffset(self) -> int:
        diff = common.Utils.from2Complement(self.immediate, 16)
        return diff*4 + 4


    def processImmediate(self, immOverride: str|None=None) -> str:
        if immOverride is not None:
            return immOverride

        if not self.descriptor.isUnsigned:
            number = common.Utils.from2Complement(self.immediate, 16)
            if number < 0:
                return f"-0x{-number:X}"
            return f"0x{number:X}"

        return f"0x{self.immediate:X}"

    def processCodeParameter(self) -> str:
        code = f"{self.instr_index >> 16}"
        lower = (self.rd << 11) | (self.sa << 6) >> 6
        if lower:
            code += f", {lower}"
        return code

    def disassembleInstruction(self, immOverride: str|None=None) -> str:
        opcode = self.getOpcodeName().lower()
        if len(self.descriptor.operands) == 0:
            return opcode

        result = opcode.ljust(InstructionConfig.OPCODE_LJUST + self.extraLjustWidthOpcode, ' ') + " "
        for i, operand in enumerate(self.descriptor.operands):
            if i != 0:
                result += ", "
            result += self.instrArgumentsCallbacks[operand](self, immOverride)

        return result

    def disassembleAsData(self) -> str:
        result = ".word".ljust(InstructionConfig.OPCODE_LJUST + self.extraLjustWidthOpcode, ' ')
        result += f" 0x{self.instr:08X}"

        return result

    def disassemble(self, immOverride: str|None=None) -> str:
        if not self.isImplemented():
            result = self.disassembleAsData()
            if InstructionConfig.UNKNOWN_INSTR_COMMENT:
                result = result.ljust(40, ' ')
                result += " # "
                result += self.disassembleInstruction(immOverride)
            return result
        return self.disassembleInstruction(immOverride)


    def mapInstrToType(self) -> str|None:
        if self.isFloatInstruction():
            if self.isDoubleFloatInstruction():
                return "f64"
            else:
                return "f32"
        # Way too general instruction to ensure the type
        # if self.uniqueId == InstructionId.LW or self.uniqueId == InstructionId.SW:
        #     return "s32"
        if self.uniqueId == InstructionId.LWU:
            return "u32"
        if self.uniqueId == InstructionId.LH or self.uniqueId == InstructionId.SH:
            return "s16"
        if self.uniqueId == InstructionId.LHU:
            return "u16"
        if self.uniqueId == InstructionId.LB or self.uniqueId == InstructionId.SB:
            return "s8"
        if self.uniqueId == InstructionId.LBU:
            return "u8"
        if self.uniqueId == InstructionId.LD or self.uniqueId == InstructionId.SD:
            return "s64"
        # if self.uniqueId == InstructionId.LDU or self.uniqueId == InstructionId.SDU:
        #     return "u64"
        return None

    def isLikelyHandwritten(self):
        if self._handwrittenCategory:
            return True

        if self.isIType() and not self.isFloatInstruction():
            if self.rs in (26, 27): # "$k0", "$k1"
                return True
            elif self.rt in (26, 27): # "$k0", "$k1"
                return True

        return False


    def __str__(self) -> str:
        return self.disassemble()

    def __repr__(self) -> str:
        return self.__str__()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

import bisect

from ... import common

from .. import symbols

from . import SectionBase


class SectionBss(SectionBase):
    def __init__(self, context: common.Context, bssVramStart: int, bssVramEnd: int, filename: str):
        super().__init__(context, bssVramStart, filename, bytearray(), common.FileSectionType.Bss)

        self.bssVramStart: int = bssVramStart
        self.bssVramEnd: int = bssVramEnd

        self.bssTotalSize: int = bssVramEnd - bssVramStart

        self.vram = bssVramStart


    def setVram(self, vram: int):
        super().setVram(vram)

        self.bssVramStart = vram
        self.bssVramEnd = vram + self.bssTotalSize

    def analyze(self):
        self.checkAndCreateFirstSymbol()

        # If something that could be a pointer found in data happens to be in the middle of this bss file's addresses space
        # Then consider it a new bss variable
        for ptr in sorted(self.context.newPointersInData):
            if ptr < self.bssVramStart:
                continue
            if ptr >= self.bssVramEnd:
                break

            contextSym = self.context.getGenericSymbol(ptr)
            if contextSym is None:
                contextSym = self.context.addSymbol(ptr, None, common.FileSectionType.Bss)
                contextSym.isAutogenerated = True
                contextSym.isDefined = True


        offsetSymbolsInSection = self.context.offsetSymbols[common.FileSectionType.Bss]
        bssSymbolOffsets: dict[int, common.ContextSymbolBase] = {offset: sym for offset, sym in offsetSymbolsInSection.items()}

        vramIndexLow = bisect.bisect(self.context.symbolsVramSorted, self.bssVramStart)
        vramIndexHigh = bisect.bisect(self.context.symbolsVramSorted, self.bssVramEnd)
        for vramIndex in range(vramIndexLow-1, vramIndexHigh-1):
            symbolVram = self.context.symbolsVramSorted[vramIndex]
            contextSym = self.context.symbols[symbolVram]
            # Mark every known symbol that happens to be in this address space as defined
            contextSym.isDefined = True
            contextSym.sectionType = common.FileSectionType.Bss

            # Needs to move this to a list because the algorithm requires to check the size of a bss variable based on the next bss variable' vram
            bssSymbolOffsets[symbolVram-self.bssVramStart] = contextSym


        sortedOffsets = sorted(bssSymbolOffsets.items())

        i = 0
        while i < len(sortedOffsets):
            symbolOffset, contextSym = sortedOffsets[i]
            symbolVram = self.bssVramStart + symbolOffset

            # Calculate the space of the bss variable
            space = self.bssTotalSize - symbolOffset
            if i + 1 < len(sortedOffsets):
                nextSymbolOffset, _ = sortedOffsets[i+1]
                if nextSymbolOffset <= self.bssTotalSize:
                    space = nextSymbolOffset - symbolOffset

            sym = symbols.SymbolBss(self.context, symbolOffset + self.inFileOffset, symbolVram, contextSym.name, space)
            sym.setCommentOffset(self.commentOffset)
            sym.analyze()
            self.symbolList.append(sym)

            i += 1



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import InstructionBase, InstructionNormal, InstructionSpecial, InstructionRegimm, InstructionCoprocessor0, InstructionCoprocessor1, InstructionCoprocessor2, InstructionNormalRsp, InstructionSpecialRsp, InstructionRegimmRsp, InstructionCoprocessor0Rsp, InstructionCoprocessor2Rsp


def wordToInstruction(word: int) -> InstructionBase:
    if ((word >> 26) & 0x3F) == 0x00:
        return InstructionSpecial(word)
    if ((word >> 26) & 0x3F) == 0x01:
        return InstructionRegimm(word)
    if ((word >> 26) & 0x3F) == 0x10:
        return InstructionCoprocessor0(word)
    if ((word >> 26) & 0x3F) == 0x11:
        return InstructionCoprocessor1(word)
    if ((word >> 26) & 0x3F) == 0x12:
        return InstructionCoprocessor2(word)
    return InstructionNormal(word)

def wordToInstructionRsp(word: int) -> InstructionBase:
    if ((word >> 26) & 0x3F) == 0x00:
        return InstructionSpecialRsp(word)
    if ((word >> 26) & 0x3F) == 0x01:
        return InstructionRegimmRsp(word)
    if ((word >> 26) & 0x3F) == 0x10:
        return InstructionCoprocessor0Rsp(word)
    # TODO?
    # if ((word >> 26) & 0x3F) == 0x11:
    #     return InstructionCoprocessor1Rsp(word)
    if ((word >> 26) & 0x3F) == 0x12:
        return InstructionCoprocessor2Rsp(word)
    return InstructionNormalRsp(word)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ... import common

from .. import instructions

from . import SymbolText


class SymbolFunction(SymbolText):
    def __init__(self, context: common.Context, inFileOffset: int, vram: int|None, name: str, instrsList: list[instructions.InstructionBase]):
        super().__init__(context, inFileOffset, vram, name, list())
        self.instructions: list[instructions.InstructionBase] = list(instrsList)

        self.pointersRemoved: bool = False

        self.localLabels: dict[int, str] = dict()
        """Branch labels found on this function.

        The key is the offset relative to the start of the function and the value is the name of the label

        If VRAM is available, then it is preferred to use `context.getGenericLabel(self.vram + branch)` to get the name of a label instead.
        """

        # TODO: this needs a better name
        self.pointersPerInstruction: dict[int, int] = dict()
        self.constantsPerInstruction: dict[int, int] = dict()
        self.branchInstructions: list[int] = list()

        # key: %hi (lui) instruction offset, value: %lo instruction offset
        self.hiToLowDict: dict[int, int] = dict()
        # key: %lo instruction offset, value: %hi (lui) instruction offset
        self.lowToHiDict: dict[int, int] = dict()

        self.luiInstructions: dict[int, instructions.InstructionBase] = dict()
        self.nonPointerLuiSet: set[int] = set()
        self.gpInstructions: dict[int, instructions.InstructionBase] = dict()

        self.pointersOffsets: set[int] = set()
        self.referencedJumpTableOffsets: set[int] = set()

        self.referencedVRams: set[int] = set()
        self.referencedConstants: set[int] = set()

        self.hasUnimplementedIntrs: bool = False

        self.isRsp: bool = False

        self.isLikelyHandwritten: bool = False

    @property
    def nInstr(self) -> int:
        return len(self.instructions)

    @property
    def vramEnd(self) -> int|None:
        if self.vram is None:
            return None
        return self.vram + self.nInstr * 4


    def _printAnalisisDebugInfo_IterInfo(self, instr: instructions.InstructionBase, register1: int|None, register2: int|None, register3: int|None, currentVram: int, trackedRegisters: dict, registersValues: dict):
        if not common.GlobalConfig.PRINT_FUNCTION_ANALYSIS_DEBUG_INFO:
            return

        print("_printAnalisisDebugInfo_IterInfo")
        print()
        print(f"vram: {currentVram:X}")
        print(instr)
        if register1 is not None:
            print(register1, instr.getRegisterName(register1))
        if register2 is not None:
            print(register2, instr.getRegisterName(register2))
        if register3 is not None:
            print(register3, instr.getRegisterName(register3))
        print(trackedRegisters)
        print({instr.getRegisterName(x): y for x, y in trackedRegisters.items()})
        # _t is shorthand of temp
        print({instr.getRegisterName(register_t): f"{vram_t:X},{offset_t:X}" for register_t, (vram_t, offset_t) in registersValues.items()})
        print()

    def _printSymbolFinderDebugInfo_DelTrackedRegister(self, instr: instructions.InstructionBase, register: int, currentVram: int|None, trackedRegisters: dict):
        if not common.GlobalConfig.PRINT_SYMBOL_FINDER_DEBUG_INFO:
            return

        if currentVram is None:
            return

        print("_printSymbolFinderDebugInfo_DelTrackedRegister")
        print()
        print(f"vram: {currentVram:X}")
        print(instr)
        print(trackedRegisters)
        print(f"deleting {register} / {instr.getRegisterName(register)}")
        print()

    def _printSymbolFinderDebugInfo_UnpairedLuis(self):
        if not common.GlobalConfig.PRINT_UNPAIRED_LUIS_DEBUG_INFO:
            return

        firstNotePrinted = False

        for instructionOffset, luiInstr in self.luiInstructions.items():
            # inFileOffset = self.inFileOffset + instructionOffset
            currentVram = self.getVramOffset(instructionOffset)
            if instructionOffset in self.nonPointerLuiSet:
                continue
            if instructionOffset in self.constantsPerInstruction:
                # print(f"{currentVram:06X} ", end="")
                # print(f"C  {self.constantsPerInstruction[instructionOffset]:8X}", luiInstr)
                pass
            else:
                if common.GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES and luiInstr.immediate < 0x4000: # filter out stuff that may not be a real symbol
                    continue
                if common.GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES and luiInstr.immediate >= 0xC000: # filter out stuff that may not be a real symbol
                    continue

                # print(f"{currentVram:06X} ", end="")
                # if instructionOffset in self.pointersPerInstruction:
                #     print(f"P  {self.pointersPerInstruction[instructionOffset]:8X}", luiInstr)
                # else:
                #     print("NO         ", luiInstr)

                if instructionOffset not in self.pointersPerInstruction:
                    if not firstNotePrinted:
                        print("_printSymbolFinderDebugInfo_UnpairedLuis")
                        print(f"func: {self.name}")
                        if self.vram is not None:
                            print(f"vram: {self.vram:08X}")
                        firstNotePrinted = True

                    print(f"{currentVram:06X} ", "NO         ", luiInstr)

        if firstNotePrinted:
            print()


    def _processBranch(self, instr: instructions.InstructionBase, instructionOffset: int, currentVram: int) -> None:
        if instr.uniqueId == instructions.InstructionId.J:
            targetBranchVram = instr.getInstrIndexAsVram()
            branch = instructionOffset + targetBranchVram - currentVram
        else:
            branch = instructionOffset + instr.getBranchOffset()
            targetBranchVram = self.getVramOffset(branch)

        self.referencedVRams.add(targetBranchVram)
        labelSym = self.context.getGenericLabel(targetBranchVram)
        if labelSym is None:
            labelSym = self.context.addBranchLabel(targetBranchVram, None, isAutogenerated=True)

        labelSym.referenceCounter += 1
        self.localLabels[branch] = labelSym.getName()
        self.branchInstructions.append(instructionOffset)

    def _processSymbol(self, luiInstr: instructions.InstructionBase|None, luiOffset: int|None, lowerInstr: instructions.InstructionBase, lowerOffset: int) -> int|None:
        # lui being None means this symbol is a $gp access
        assert ((luiInstr is None and luiOffset is None) or (luiInstr is not None and luiOffset is not None))

        if lowerOffset in self.pointersPerInstruction:
            # This %lo has been processed already
            return self.pointersPerInstruction[lowerOffset]

        if luiInstr is None and common.GlobalConfig.GP_VALUE is None:
            return None

        if luiInstr is not None:
            upperHalf = luiInstr.immediate << 16
        else:
            assert common.GlobalConfig.GP_VALUE is not None
            upperHalf = common.GlobalConfig.GP_VALUE

        lowerHalf = common.Utils.from2Complement(lowerInstr.immediate, 16)
        address = upperHalf + lowerHalf
        if address in self.context.bannedSymbols:
            return None

        # filter out stuff that may not be a real symbol
        filterOut = common.GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES and upperHalf < 0x40000000
        filterOut |= common.GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES and upperHalf >= 0xC0000000
        if filterOut:
            if common.GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_CONSTANTS:
                # Let's pretend this value is a constant
                constant = address
                self.referencedConstants.add(constant)

                self.constantsPerInstruction[lowerOffset] = constant
                if luiOffset is not None:
                    self.constantsPerInstruction[luiOffset] = constant

                    self.hiToLowDict[luiOffset] = lowerOffset
                    self.lowToHiDict[lowerOffset] = luiOffset
            return None

        patchedAddress = address
        patch = self.context.getLoPatch(lowerInstr.vram)
        if patch is not None:
            patchedAddress = patch

        self.referencedVRams.add(patchedAddress)
        contextSym = self.context.getGenericSymbol(patchedAddress)
        if contextSym is None:
            if common.GlobalConfig.ADD_NEW_SYMBOLS:
                contextSym = self.context.addSymbol(patchedAddress, None, isAutogenerated=True)
                instrType = lowerInstr.mapInstrToType()
                if instrType is not None:
                    contextSym.setTypeIfUnset(instrType)
                contextSym.referenceCounter += 1
        else:
            contextSym.referenceCounter += 1

        if lowerOffset not in self.pointersPerInstruction:
            self.pointersPerInstruction[lowerOffset] = address
        if luiOffset is not None:
            if luiOffset not in self.pointersPerInstruction:
                self.pointersPerInstruction[luiOffset] = address

            self.hiToLowDict[luiOffset] = lowerOffset
            self.lowToHiDict[lowerOffset] = luiOffset
        else:
            self.gpInstructions[lowerOffset] = lowerInstr

        return address

    def _processConstant(self, luiInstr: instructions.InstructionBase, luiOffset: int, lowerInstr: instructions.InstructionBase, lowerOffset: int) -> int|None:
        upperHalf = luiInstr.immediate << 16
        lowerHalf = lowerInstr.immediate
        constant = upperHalf | lowerHalf

        self.referencedConstants.add(constant)

        self.constantsPerInstruction[lowerOffset] = constant
        self.constantsPerInstruction[luiOffset] = constant

        self.hiToLowDict[luiOffset] = lowerOffset
        self.lowToHiDict[lowerOffset] = luiOffset

        return constant

    def _removeRegisterFromTrackers(self, instr: instructions.InstructionBase, prevInstr: instructions.InstructionBase|None, currentVram: int|None, trackedRegisters: dict, trackedRegistersAll: dict, registersValues: dict, wasRegisterValuesUpdated: bool):
        shouldRemove = False
        register = 0

        if prevInstr is not None and prevInstr.isBranchLikely():
            return

        if not instr.isFloatInstruction():
            if instr.isRType() or (instr.isBranch() and isinstance(instr, instructions.InstructionNormal)):
                # $at is a one-use register
                at = -1
                if instr.rs == 1:
                    at = instr.rs
                elif instr.rt == 1:
                    at = instr.rt

                if at in trackedRegistersAll:
                    otherInstrIndex = trackedRegistersAll[at]
                    otherInstr = self.instructions[otherInstrIndex]
                    if otherInstr.uniqueId == instructions.InstructionId.LUI:
                        self.nonPointerLuiSet.add(otherInstrIndex*4)
                    shouldRemove = True
                    register = at

            if instr.uniqueId != instructions.InstructionId.LUI and instr.modifiesRt():
                shouldRemove = True
                register = instr.rt

            if instr.modifiesRd():
                shouldRemove = True
                register = instr.rd

                # Usually array offsets use an ADDU to add the index of the array
                if instr.uniqueId == instructions.InstructionId.ADDU:
                    if instr.rd != instr.rs and instr.rd != instr.rt:
                        shouldRemove = True
                    else:
                        shouldRemove = False

        else:
            if instr.uniqueId in (instructions.InstructionId.MTC1, instructions.InstructionId.DMTC1, instructions.InstructionId.CTC1):
                # IDO usually use a register as a temp when loading a constant value
                # into the float coprocessor, after that IDO never re-uses the value
                # in that register for anything else
                shouldRemove = True
                register = instr.rt

        if shouldRemove:
            if register in trackedRegisters:
                self._printSymbolFinderDebugInfo_DelTrackedRegister(instr, register, currentVram, trackedRegisters)
                del trackedRegisters[register]
            if register in trackedRegistersAll:
                del trackedRegistersAll[register]
            if not wasRegisterValuesUpdated:
                if register in registersValues:
                    del registersValues[register]

    def _tryToSetSymbolType(self, instr: instructions.InstructionBase, instructionOffset: int, registersValues: dict[int, tuple[int, int]]):
        instrType = instr.mapInstrToType()
        if instrType is None or instr.rs not in registersValues:
            return

        address, loInstructionOffset = registersValues[instr.rs]
        if instructionOffset != loInstructionOffset:
            loInstr = self.instructions[loInstructionOffset//4]
            if loInstr.uniqueId != instructions.InstructionId.ADDIU:
                # if the instruction used to load this value wasn't an ADDIU
                # then the register has the value pointed by this address
                return

        contextSym = self.context.getSymbol(address, tryPlusOffset=False)
        if contextSym is not None:
            contextSym.setTypeIfUnset(instrType)

    def _symbolFinder(self, instr: instructions.InstructionBase, prevInstr: instructions.InstructionBase, instructionOffset: int, trackedRegisters: dict[int, int], trackedRegistersAll: dict[int, int], registersValues: dict[int, tuple[int, int]]):
        if instr.uniqueId == instructions.InstructionId.LUI:
            if not prevInstr.isBranchLikely():
                # If the previous instructions is a branch likely, then nulify
                # the effects of this instruction for future analysis
                trackedRegisters[instr.rt] = instructionOffset//4
            trackedRegistersAll[instr.rt] = instructionOffset//4
            return False

        if instr.uniqueId == instructions.InstructionId.ORI:
            # Constants
            rs = instr.rs
            if rs in trackedRegistersAll:
                luiOffset = trackedRegistersAll[rs] * 4
                luiInstr = self.instructions[luiOffset//4]
                constant = self._processConstant(luiInstr, luiOffset, instr, instructionOffset)
                if constant is not None:
                    registersValues[instr.rt] = (constant, instructionOffset)
                    return True
            return False

        if instr.uniqueId not in (instructions.InstructionId.ANDI, instructions.InstructionId.XORI, instructions.InstructionId.CACHE, instructions.InstructionId.SLTI, instructions.InstructionId.SLTIU):
            rs = instr.rs
            if rs in trackedRegisters:
                luiInstr = self.instructions[trackedRegisters[rs]]
                address = self._processSymbol(luiInstr, trackedRegisters[rs]*4, instr, instructionOffset)
                if address is not None:
                    registersValues[instr.rt] = (address, instructionOffset)
                    return True
            elif rs == 28: # $gp
                address = self._processSymbol(None, None, instr, instructionOffset)
                if address is not None:
                    registersValues[instr.rt] = (address, instructionOffset)
                    return True

        return False


    def _lookAheadSymbolFinder(self, instr: instructions.InstructionBase, instructionOffset: int, trackedRegistersOriginal: dict[int, int], trackedRegistersAllOriginal: dict[int, int], registersValuesOriginal: dict[int, tuple[int, int]]):
        trackedRegisters = dict(trackedRegistersOriginal)
        trackedRegistersAll = dict(trackedRegistersAllOriginal)
        registersValues = dict(registersValuesOriginal)

        lastInstr = self.instructions[instructionOffset//4 - 1]
        if not lastInstr.isBranch():
            return

        branchOffset = lastInstr.getBranchOffset() - 4
        branch = instructionOffset + branchOffset
        # don't check negative branches (loops) or branches outside this function
        if branchOffset <= 0 or branch//4 >= len(self.instructions):
            return

        self._removeRegisterFromTrackers(instr, None, None, trackedRegisters, trackedRegistersAll, registersValues, False)

        pairedLoFound = False
        i = 0
        while True:
            if branch//4 >= len(self.instructions):
                return

            if i >= 5:
                if instr.uniqueId == instructions.InstructionId.LUI:
                    # Continue searching until we find the corresponding lo instruction for this LUI
                    if pairedLoFound:
                        return
                else:
                    # Only check the 5 next instructions in the target branch for non LUI instructions
                    return

            wasRegisterValuesUpdated = False
            prevTargetInstr = self.instructions[branch//4 - 1]
            targetInstr = self.instructions[branch//4]

            # Usually array offsets use an ADDU to add the index of the array
            if targetInstr.uniqueId == instructions.InstructionId.ADDU and not prevTargetInstr.isBranchLikely():
                if targetInstr.rd == targetInstr.rs or targetInstr.rd == targetInstr.rt:
                    branch += 4
                    i += 1
                    continue

            if targetInstr.isIType():
                if self._symbolFinder(targetInstr, prevTargetInstr, branch, trackedRegisters, trackedRegistersAll, registersValues):
                    wasRegisterValuesUpdated = True
                    if instr.uniqueId == instructions.InstructionId.LUI and targetInstr.rs == instr.rt:
                        pairedLoFound = True
                self._tryToSetSymbolType(targetInstr, branch, registersValues)

            if prevTargetInstr.uniqueId == instructions.InstructionId.B or (prevTargetInstr.uniqueId == instructions.InstructionId.BEQ and prevTargetInstr.rt == 0 and prevTargetInstr.rs == 0):
                # TODO: Consider following branches
                # self._lookAheadSymbolFinder(targetInstr, branch, trackedRegisters, trackedRegistersAll, registersValues)
                return
            if prevTargetInstr.isJType():
                return
            if prevTargetInstr.isJump():
                return

            self._removeRegisterFromTrackers(targetInstr, prevTargetInstr, None, trackedRegisters, trackedRegistersAll, registersValues, wasRegisterValuesUpdated)

            branch += 4
            i += 1

    def _processElfRelocSymbols(self):
        if len(self.context.relocSymbols[common.FileSectionType.Text]) == 0:
            return

        # Process reloc symbols (probably from a .elf file)
        instructionOffset = 0
        inFileOffset = self.inFileOffset
        for instr in self.instructions:
            relocSymbol = self.context.getRelocSymbol(inFileOffset, common.FileSectionType.Text)
            if relocSymbol is not None:
                if relocSymbol.name.startswith("."):
                    sectType = common.FileSectionType.fromStr(relocSymbol.name)

                    if instructionOffset in self.pointersPerInstruction:
                        if instructionOffset in self.referencedJumpTableOffsets:
                            # Jump tables
                            addressOffset = self.pointersPerInstruction[instructionOffset]
                            if relocSymbol.name != ".rodata":
                                common.Utils.eprint(f"Warning. Jumptable referenced in reloc does not have '.rodata' as its name")
                            contextOffsetSym = self.context.addOffsetJumpTable(addressOffset, sectType)
                            contextOffsetSym.referenceCounter += 1
                            relocSymbol.name = contextOffsetSym.name
                            self.pointersPerInstruction[instructionOffset] = 0
                            if instructionOffset in self.lowToHiDict:
                                luiOffset = self.lowToHiDict[instructionOffset]
                                otherReloc = self.context.getRelocSymbol(self.inFileOffset+luiOffset, common.FileSectionType.Text)
                                if otherReloc is not None:
                                    otherReloc.name = relocSymbol.name
                                    self.pointersPerInstruction[luiOffset] = 0
                        else:
                            addressOffset = self.pointersPerInstruction[instructionOffset]
                            relocName = f"{relocSymbol.name}_{addressOffset:06X}"
                            # print(relocName, addressOffset, instr)
                            contextOffsetSym = common.ContextOffsetSymbol(addressOffset, relocName, sectType)
                            self.context.offsetSymbols[sectType][addressOffset] = contextOffsetSym
                            relocSymbol.name = relocName
                            self.pointersPerInstruction[instructionOffset] = 0
            inFileOffset += 4
            instructionOffset += 4


    def analyze(self):
        if not common.GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS and self.hasUnimplementedIntrs:
            if self.vram is not None:
                offset = 0
                for instr in self.instructions:
                    currentVram = self.getVramOffset(offset)
                    contextSym = self.context.getSymbol(currentVram, False)
                    if contextSym is not None:
                        contextSym.isDefined = True

                    offset += 4
            return

        # Search for LUI instructions first
        instructionOffset = 0
        for instr in self.instructions:
            if instr.uniqueId == instructions.InstructionId.LUI:
                self.luiInstructions[instructionOffset] = instr
            if instructionOffset > 0:
                prevInstr = self.instructions[instructionOffset//4 - 1]
                if prevInstr.isJType() or prevInstr.isJump():
                    self.nonPointerLuiSet.add(instructionOffset)
            instructionOffset += 4

        trackedRegisters: dict[int, int] = dict()
        trackedRegistersAll: dict[int, int] = dict()
        # key: register, value: (vram, offset of instruction which set this value)
        registersValues: dict[int, tuple[int, int]] = dict()

        instructionOffset = 0
        for instr in self.instructions:
            currentVram = self.getVramOffset(instructionOffset)
            wasRegisterValuesUpdated = False
            self.isLikelyHandwritten |= instr.uniqueId in instructions.InstructionsNotEmitedByIDO
            prevInstr = self.instructions[instructionOffset//4 - 1]

            self._printAnalisisDebugInfo_IterInfo(instr, instr.rs, instr.rt, instr.rd, currentVram, trackedRegisters, registersValues)

            if not self.isLikelyHandwritten:
                self.isLikelyHandwritten = instr.isLikelyHandwritten()

            if not common.GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS and not instr.isImplemented():
                # Abort analysis
                self.hasUnimplementedIntrs = True
                return

            if instr.isBranch() or (common.GlobalConfig.TREAT_J_AS_UNCONDITIONAL_BRANCH and instr.uniqueId == instructions.InstructionId.J):
                self._processBranch(instr, instructionOffset, currentVram)

            elif instr.isJType():
                target = instr.getInstrIndexAsVram()
                if not self.isRsp:
                    if target >= 0x84000000:
                        # RSP address space?
                        self.isLikelyHandwritten = True
                funcSym = self.context.getFunction(target)
                if funcSym is None:
                    if instr.uniqueId == instructions.InstructionId.J and not self.isRsp:
                        funcSym = self.context.addFakeFunction(target, f".L{target:08X}", isAutogenerated=True)
                    else:
                        funcSym = self.context.addFunction(target, f"func_{target:08X}", isAutogenerated=True)
                funcSym.referenceCounter += 1
                self.pointersPerInstruction[instructionOffset] = target

            # symbol finder
            elif instr.isIType():
                if self._symbolFinder(instr, prevInstr, instructionOffset, trackedRegisters, trackedRegistersAll, registersValues):
                    wasRegisterValuesUpdated = True
                self._tryToSetSymbolType(instr, instructionOffset, registersValues)

            elif instr.uniqueId == instructions.InstructionId.JR:
                rs = instr.rs
                if rs != 31: # $ra
                    if rs in registersValues:
                        # print(instructionOffset, rs, trackedRegisters, trackedRegistersAll, registersValues, self.pointersPerInstruction)
                        address, jmptblSeterOffset = registersValues[rs]
                        self.referencedJumpTableOffsets.add(jmptblSeterOffset)
                        self.referencedVRams.add(address)
                        self.context.addJumpTable(address, isAutogenerated=True)

            self._removeRegisterFromTrackers(instr, prevInstr, currentVram, trackedRegisters, trackedRegistersAll, registersValues, wasRegisterValuesUpdated)

            # look-ahead symbol finder
            self._lookAheadSymbolFinder(instr, instructionOffset, trackedRegisters, trackedRegistersAll, registersValues)

            instructionOffset += 4

        self._printSymbolFinderDebugInfo_UnpairedLuis()

        self._processElfRelocSymbols()

        for instr in self.instructions:
            instr.inHandwrittenFunction = self.isLikelyHandwritten


    def countDiffOpcodes(self, other: SymbolFunction) -> int:
        result = 0
        for i in range(min(self.nInstr, other.nInstr)):
            if not self.instructions[i].sameOpcode(other.instructions[i]):
                result += 1
        return result

    def countSameOpcodeButDifferentArguments(self, other: SymbolFunction) -> int:
        result = 0
        for i in range(min(self.nInstr, other.nInstr)):
            instr1 = self.instructions[i]
            instr2 = other.instructions[i]
            if instr1.sameOpcodeButDifferentArguments(instr2):
                result += 1
        return result

    def blankOutDifferences(self, other_func: SymbolFunction) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = False

        for i in range(min(self.nInstr, other_func.nInstr)):
            instr1 = self.instructions[i]
            instr2 = other_func.instructions[i]
            if instr1.sameOpcodeButDifferentArguments(instr2):
                instr1.blankOut()
                instr2.blankOut()
                was_updated = True

        return was_updated

    def removePointers(self) -> bool:
        if not common.GlobalConfig.REMOVE_POINTERS:
            return False

        was_updated = False

        for instructionOffset in self.pointersPerInstruction:
            self.instructions[instructionOffset//4].blankOut()
        was_updated = len(self.pointersPerInstruction) > 0 or was_updated

        for fileOffset in self.pointersOffsets:
            index = (fileOffset - self.inFileOffset)//4
            if index < 0:
                continue
            if index >= self.nInstr:
                continue
            self.instructions[index].blankOut()

        if common.GlobalConfig.IGNORE_BRANCHES:
            for instructionOffset in self.branchInstructions:
                self.instructions[instructionOffset//4].blankOut()
            was_updated = len(self.branchInstructions) > 0 or was_updated

        self.pointersRemoved = True

        return was_updated

    def removeTrailingNops(self) -> bool:
        was_updated = False
        first_nop = self.nInstr

        for i in range(self.nInstr-1, 0-1, -1):
            instr = self.instructions[i]
            if instr.uniqueId != instructions.InstructionId.NOP:
                if instr.uniqueId == instructions.InstructionId.JR and instr.rs == 31: #$ra
                    first_nop += 1
                break
            first_nop = i

        if first_nop < self.nInstr:
            was_updated = True
            del self.instructions[first_nop:]
        return was_updated


    def generateHiLoStr(self, instr: instructions.InstructionBase, symName: str) -> str:
        if instr.uniqueId == instructions.InstructionId.LUI:
            return f"%hi({symName})"

        # $gp
        if instr.rs == 28:
            # $gp
            if instr.rt != 28 or not instr.modifiesRt():
                return f"%gp_rel({symName})"

        return f"%lo({symName})"

    def getImmOverrideForInstruction(self, instr: instructions.InstructionBase, instructionOffset: int) -> str|None:
        if len(self.context.relocSymbols[self.sectionType]) > 0:
            # Check possible symbols using reloc information (probably from a .o elf file)
            possibleImmOverride = self.context.getRelocSymbol(self.inFileOffset + instructionOffset, self.sectionType)
            if possibleImmOverride is not None:
                auxOverride = possibleImmOverride.name
                if instr.isIType():
                    if instructionOffset in self.pointersPerInstruction:
                        addressOffset = self.pointersPerInstruction[instructionOffset]
                        auxOverride = possibleImmOverride.getNamePlusOffset(addressOffset)

                    auxOverride = self.generateHiLoStr(instr, auxOverride)
                return auxOverride

        if instr.isBranch():
            if not common.GlobalConfig.IGNORE_BRANCHES:
                if instr.uniqueId == instructions.InstructionId.J:
                    targetBranchVram = instr.getInstrIndexAsVram()
                    branch = instructionOffset + targetBranchVram - self.getVramOffset(instructionOffset)
                else:
                    branch = instructionOffset + instr.getBranchOffset()
                    targetBranchVram = self.getVramOffset(branch)
                labelSymbol = self.context.getGenericLabel(targetBranchVram)
                if labelSymbol is not None:
                    return labelSymbol.name

                # in case we don't have access to vram or this label was not in context
                if branch in self.localLabels:
                    return self.localLabels[branch]

        elif instr.isIType():
            if not self.pointersRemoved and instructionOffset in self.pointersPerInstruction:
                address = self.pointersPerInstruction[instructionOffset]

                instrVram = instr.vram
                if instr.uniqueId == instructions.InstructionId.LUI:
                    # we need to get the address of the lo instruction to get the patch
                    if instructionOffset in self.hiToLowDict:
                        loInstr = self.instructions[self.hiToLowDict[instructionOffset] // 4]
                        instrVram = loInstr.vram

                # Check for user-defined symbol patches
                patchedAddress = self.context.getLoPatch(instrVram)
                if patchedAddress is not None:
                    symbol = self.context.getGenericSymbol(patchedAddress, True, False)
                else:
                    symbol = self.context.getGenericSymbol(address, True)

                if symbol is not None:
                    return self.generateHiLoStr(instr, symbol.getSymbolPlusOffset(address))

            elif instructionOffset in self.constantsPerInstruction:
                constant = self.constantsPerInstruction[instructionOffset]

                symbol = self.context.getConstant(constant)
                if symbol is not None:
                    return self.generateHiLoStr(instr, symbol.name)

                if instr.uniqueId == instructions.InstructionId.LUI:
                    loInstr = self.instructions[self.hiToLowDict[instructionOffset] // 4]
                    if loInstr.uniqueId == instructions.InstructionId.ORI:
                        return f"(0x{constant:X} >> 16)"
                elif instr.uniqueId == instructions.InstructionId.ORI:
                    return f"(0x{constant:X} & 0xFFFF)"
                return self.generateHiLoStr(instr, f"0x{constant:X}")

            elif instr.uniqueId == instructions.InstructionId.LUI:
                return f"(0x{instr.immediate<<16:X} >> 16)"

        elif instr.isJType():
            possibleOverride = self.context.getAnySymbol(instr.getInstrIndexAsVram())
            if possibleOverride is not None:
                return possibleOverride.name

        return None

    def getLabelForOffset(self, instructionOffset: int) -> str:
        if not common.GlobalConfig.IGNORE_BRANCHES and instructionOffset != 0:
            # Skip over this function to avoid duplication

            currentVram = self.getVramOffset(instructionOffset)
            labelSym = self.context.getGenericLabel(currentVram)
            if labelSym is None and len(self.context.offsetJumpTablesLabels) > 0:
                labelSym = self.context.getOffsetGenericLabel(self.inFileOffset+instructionOffset, common.FileSectionType.Text)
            if labelSym is None and len(self.context.offsetSymbols[self.sectionType]) > 0:
                labelSym = self.context.getOffsetSymbol(self.inFileOffset+instructionOffset, common.FileSectionType.Text)

            if labelSym is not None:
                labelSym.isDefined = True
                labelSym.sectionType = self.sectionType
                if labelSym.type == common.SymbolSpecialType.function or labelSym.type == common.SymbolSpecialType.jumptablelabel:
                    return labelSym.getSymbolLabel() + common.GlobalConfig.LINE_ENDS
                return labelSym.getName() + ":" + common.GlobalConfig.LINE_ENDS

            if instructionOffset in self.localLabels:
                return self.localLabels[instructionOffset] + ":" + common.GlobalConfig.LINE_ENDS
        return ""


    def disassemble(self) -> str:
        output = ""

        if not common.GlobalConfig.DISASSEMBLE_UNKNOWN_INSTRUCTIONS:
            if self.hasUnimplementedIntrs:
                return self.disassembleAsData()

        if self.isLikelyHandwritten:
            output += "/* Handwritten function */" + common.GlobalConfig.LINE_ENDS

        output += self.getLabel()

        if common.GlobalConfig.ASM_TEXT_ENT_LABEL:
            output += f"{common.GlobalConfig.ASM_TEXT_ENT_LABEL} {self.name}" + common.GlobalConfig.LINE_ENDS

        if common.GlobalConfig.ASM_TEXT_FUNC_AS_LABEL:
            output += f"{self.name}:" + common.GlobalConfig.LINE_ENDS

        wasLastInstABranch = False
        instructionOffset = 0
        for instr in self.instructions:
            immOverride = self.getImmOverrideForInstruction(instr, instructionOffset)
            comment = self.generateAsmLineComment(instructionOffset, instr.instr)

            if wasLastInstABranch:
                instr.extraLjustWidthOpcode -= 1
                comment += " "

            line = instr.disassemble(immOverride)

            if wasLastInstABranch:
                instr.extraLjustWidthOpcode += 1

            label = self.getLabelForOffset(instructionOffset)
            output += f"{label}{comment}  {line}" + common.GlobalConfig.LINE_ENDS

            wasLastInstABranch = instr.isBranch() or instr.isJump()
            instructionOffset += 4

        if common.GlobalConfig.ASM_TEXT_END_LABEL:
            output += f"{common.GlobalConfig.ASM_TEXT_END_LABEL} {self.name}" + common.GlobalConfig.LINE_ENDS

        return output

    def disassembleAsData(self) -> str:
        self.words = [instr.instr for instr in self.instructions]
        return super().disassembleAsData()
